import logging
import uuid

from collections import defaultdict
from contextlib import contextmanager

# top-level include is dangerous in terms of incurring circular deps
from dagster import (
    DagsterInvariantViolationError,
    DependencyDefinition,
    ModeDefinition,
    PipelineDefinition,
    SolidDefinition,
    SolidInvocation,
    SystemStorageData,
    check,
    execute_pipeline,
    lambda_solid,
)
from dagster.core.definitions.logger import LoggerDefinition
from dagster.core.definitions.resource import ScopedResourcesBuilder
from dagster.core.execution.api import RunConfig, scoped_pipeline_context
from dagster.core.execution.context_creation_pipeline import (
    create_log_manager,
    construct_pipeline_execution_context,
    create_context_creation_data,
)
from dagster.core.utility_solids import define_stub_solid
from dagster.core.storage.intermediates_manager import InMemoryIntermediatesManager
from dagster.core.storage.file_manager import LocalFileManager
from dagster.core.storage.runs import InMemoryRunStorage

# pylint: disable=unused-import
from .temp_file import (
    get_temp_file_handle,
    get_temp_file_handle_with_data,
    get_temp_file_name,
    get_temp_file_name_with_data,
    get_temp_file_names,
)


def create_test_pipeline_execution_context(
    logger_defs=None, scoped_resources_builder=None, tags=None, run_config_loggers=None
):
    run_id = str(uuid.uuid4())
    loggers = check.opt_dict_param(
        logger_defs, 'logger_defs', key_type=str, value_type=LoggerDefinition
    )
    mode_def = ModeDefinition(logger_defs=loggers)
    pipeline_def = PipelineDefinition(
        name='test_legacy_context', solid_defs=[], mode_defs=[mode_def]
    )
    run_config_loggers = check.opt_list_param(
        run_config_loggers, 'run_config_loggers', of_type=logging.Logger
    )
    run_config = RunConfig(run_id, tags=tags, loggers=run_config_loggers)
    environment_dict = {'loggers': {key: {} for key in loggers}}
    creation_data = create_context_creation_data(pipeline_def, environment_dict, run_config)
    log_manager = create_log_manager(creation_data)

    scoped_resources_builder = check.opt_inst_param(
        scoped_resources_builder,
        'scoped_resources_builder',
        ScopedResourcesBuilder,
        default=ScopedResourcesBuilder(),
    )
    return construct_pipeline_execution_context(
        context_creation_data=creation_data,
        scoped_resources_builder=scoped_resources_builder,
        system_storage_data=SystemStorageData(
            run_storage=InMemoryRunStorage(),
            intermediates_manager=InMemoryIntermediatesManager(),
            file_manager=LocalFileManager.for_run_id(run_id),
        ),
        log_manager=log_manager,
    )


def _dep_key_of(solid):
    return SolidInvocation(solid.definition.name, solid.name)


def build_pipeline_with_input_stubs(pipeline_def, inputs):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.dict_param(inputs, 'inputs', key_type=str, value_type=dict)

    deps = defaultdict(dict)
    for solid_name, dep_dict in pipeline_def.dependencies.items():
        for input_name, dep in dep_dict.items():
            deps[solid_name][input_name] = dep

    stub_solid_defs = []

    for solid_name, input_dict in inputs.items():
        if not pipeline_def.has_solid_named(solid_name):
            raise DagsterInvariantViolationError(
                (
                    'You are injecting an input value for solid {solid_name} '
                    'into pipeline {pipeline_name} but that solid was not found'
                ).format(solid_name=solid_name, pipeline_name=pipeline_def.name)
            )

        solid = pipeline_def.solid_named(solid_name)
        for input_name, input_value in input_dict.items():
            stub_solid_def = define_stub_solid(
                '__stub_{solid_name}_{input_name}'.format(
                    solid_name=solid_name, input_name=input_name
                ),
                input_value,
            )
            stub_solid_defs.append(stub_solid_def)
            deps[_dep_key_of(solid)][input_name] = DependencyDefinition(stub_solid_def.name)

    return PipelineDefinition(
        name=pipeline_def.name + '_stubbed',
        solid_defs=pipeline_def.solid_defs + stub_solid_defs,
        mode_defs=pipeline_def.mode_definitions,
        dependencies=deps,
    )


def execute_solids_within_pipeline(
    pipeline_def, solid_names, inputs=None, environment_dict=None, run_config=None
):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.list_param(solid_names, 'solid_names', of_type=str)
    inputs = check.opt_dict_param(inputs, 'inputs', key_type=str, value_type=dict)
    environment_dict = check.opt_dict_param(environment_dict, 'environment_dict')
    run_config = check.opt_inst_param(run_config, 'run_config', RunConfig)

    sub_pipeline = pipeline_def.build_sub_pipeline(solid_names)
    stubbed_pipeline = build_pipeline_with_input_stubs(sub_pipeline, inputs)
    result = execute_pipeline(stubbed_pipeline, environment_dict, run_config)

    return {sr.solid.name: sr for sr in result.solid_result_list}


def execute_solid_within_pipeline(
    pipeline_def, solid_name, inputs=None, environment_dict=None, run_config=None
):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.str_param(solid_name, 'solid_name')
    inputs = check.opt_dict_param(inputs, 'inputs', key_type=str)
    environment_dict = check.opt_dict_param(environment_dict, 'environment')
    run_config = check.opt_inst_param(run_config, 'run_config', RunConfig)

    return execute_solids_within_pipeline(
        pipeline_def,
        [solid_name],
        {solid_name: inputs} if inputs else None,
        environment_dict,
        run_config,
    )[solid_name]


@contextmanager
def yield_empty_pipeline_context(run_id=None):
    with scoped_pipeline_context(PipelineDefinition([]), {}, RunConfig(run_id=run_id)) as context:
        yield context


def execute_solid(
    solid_def, mode_def=None, input_values=None, environment_dict=None, run_config=None
):
    '''
    Independently execute an individual solid without having to specify a pipeline. This
    also allows one to directly pass in in-memory input values. This is very
    useful for unit test cases.

    '''
    check.inst_param(solid_def, 'solid_def', SolidDefinition)
    check.opt_inst_param(mode_def, 'mode_def', ModeDefinition)
    input_values = check.opt_dict_param(input_values, 'input_values', key_type=str)

    solid_defs = [solid_def]

    def create_value_solid(input_name, input_value):
        @lambda_solid(name=input_name)
        def input_solid():
            return input_value

        return input_solid

    dependencies = defaultdict(dict)

    for input_name, input_value in input_values.items():
        dependencies[solid_def.name][input_name] = DependencyDefinition(input_name)
        solid_defs.append(create_value_solid(input_name, input_value))

    return execute_pipeline(
        PipelineDefinition(
            name='emphemeral_{}_solid_pipeline'.format(solid_def.name),
            solid_defs=solid_defs,
            dependencies=dependencies,
            mode_defs=[mode_def] if mode_def else None,
        ),
        environment_dict=environment_dict,
        run_config=run_config,
    ).result_for_solid(solid_def.name)



from .config import ConfigMapping

from .dependency import (
    DependencyDefinition,
    MultiDependencyDefinition,
    Solid,
    SolidHandle,
    SolidInputHandle,
    SolidInvocation,
    SolidOutputHandle,
)

from .decorators import solid, lambda_solid, composite_solid, pipeline

from .environment_schema import (
    EnvironmentSchema,
    create_environment_schema,
    create_environment_type,
)

from .expectation import IOExpectationDefinition, DagsterIOExpectationFailedError

from .events import (
    EventMetadataEntry,
    TextMetadataEntryData,
    UrlMetadataEntryData,
    PathMetadataEntryData,
    JsonMetadataEntryData,
    ExpectationResult,
    Failure,
    Materialization,
    Output,
    TypeCheck,
)

from .handle import ExecutionTargetHandle, LoaderEntrypoint

from .input import InputDefinition, InputMapping

from .logger import LoggerDefinition, logger

from .output import OutputDefinition, OutputMapping

from .resource import ResourceDefinition, resource

from .mode import ModeDefinition

from .repository import RepositoryDefinition

from .pipeline import PipelineDefinition

from .system_storage import SystemStorageDefinition, SystemStorageData, system_storage

from .container import solids_in_topological_order, create_execution_structure, IContainSolids

from .solid import SolidDefinition, ISolidDefinition, CompositeSolidDefinition

from .preset import PresetDefinition



import itertools
import os
import shutil
import tempfile

from contextlib import contextmanager

from dagster import check
from dagster.core.storage.file_manager import LocalFileHandle


def _unlink_swallow_errors(path):
    check.str_param(path, 'path')
    try:
        os.unlink(path)
    except Exception:  # pylint: disable=broad-except
        pass


@contextmanager
def get_temp_file_handle_with_data(data):
    with get_temp_file_name_with_data(data) as temp_file:
        yield LocalFileHandle(temp_file)


@contextmanager
def get_temp_file_name_with_data(data):
    with get_temp_file_name() as temp_file:
        with open(temp_file, 'wb') as ff:
            ff.write(data)

        yield temp_file


@contextmanager
def get_temp_file_handle():
    with get_temp_file_name() as temp_file:
        yield LocalFileHandle(temp_file)


@contextmanager
def get_temp_file_name():
    temp_file_name = tempfile.mkstemp()[1]
    try:
        yield temp_file_name
    finally:
        _unlink_swallow_errors(temp_file_name)


@contextmanager
def get_temp_file_names(number):
    check.int_param(number, 'number')

    temp_file_names = list()
    for _ in itertools.repeat(None, number):
        temp_file_name = tempfile.mkstemp()[1]
        temp_file_names.append(temp_file_name)

    try:
        yield tuple(temp_file_names)
    finally:
        for temp_file_name in temp_file_names:
            _unlink_swallow_errors(temp_file_name)


@contextmanager
def get_temp_dir():
    temp_dir = None
    try:
        temp_dir = tempfile.mkdtemp()
        yield temp_dir
    finally:
        if temp_dir:
            shutil.rmtree(temp_dir)



from collections import namedtuple

from dagster import check

from dagster.core.types import Field, String
from dagster.core.types.field_utils import check_user_facing_opt_field_param
from .config import resolve_config_field


class ResourceDefinition(object):
    '''Resources are pipeline-scoped ways to make external resources (like database connections)
    available to solids during pipeline execution and clean up after execution resolves.

    Args:
        resource_fn (Callable[[InitResourceContext], Any]):
            User provided function to instantiate the resource. This resource will be available to
            solids via ``context.resources``
        config_field (Field):
            The type for the configuration data for this resource, passed to ``resource_fn`` via
            ``init_context.resource_config``
        description (str)
    '''

    def __init__(self, resource_fn, config_field=None, description=None):
        self.resource_fn = check.callable_param(resource_fn, 'resource_fn')
        self.config_field = check_user_facing_opt_field_param(
            config_field, 'config_field', 'of a ResourceDefinition or @resource'
        )
        self.description = check.opt_str_param(description, 'description')

    @staticmethod
    def none_resource(description=None):
        return ResourceDefinition.hardcoded_resource(value=None, description=description)

    @staticmethod
    def hardcoded_resource(value, description=None):
        return ResourceDefinition(resource_fn=lambda _init_context: value, description=description)

    @staticmethod
    def string_resource(description=None):
        return ResourceDefinition(
            resource_fn=lambda init_context: init_context.resource_config,
            config_field=Field(String),
            description=description,
        )


def resource(config=None, config_field=None, description=None):
    '''A decorator for creating a resource. The decorated function will be used as the
    resource_fn in a ResourceDefinition.

    Args:
        config (Dict[str, Field]):
            The schema for the configuration data to be made available to the resource_fn
        config_field (Field):
            Used in the rare case of a top level config type other than a dictionary.

            Only one of config or config_field can be provided.
        description(str)
    '''

    # This case is for when decorator is used bare, without arguments.
    # E.g. @resource versus @resource()
    if callable(config):
        return ResourceDefinition(resource_fn=config)

    def _wrap(resource_fn):
        return ResourceDefinition(
            resource_fn, resolve_config_field(config_field, config, '@resource'), description
        )

    return _wrap


class ScopedResourcesBuilder(namedtuple('ScopedResourcesBuilder', 'resource_instance_dict')):
    '''There are concepts in the codebase (e.g. solids, system storage) that receive
    only the resources that they have specified in required_resource_keys.
    ScopedResourcesBuilder is responsible for dynamically building a class with
    only those required resources and returning an instance of that class.'''

    def __new__(cls, resource_instance_dict=None):
        return super(ScopedResourcesBuilder, cls).__new__(
            cls,
            resource_instance_dict=check.opt_dict_param(
                resource_instance_dict, 'resource_instance_dict', key_type=str
            ),
        )

    def build(self, mapper_fn=None, required_resource_keys=None):
        '''We dynamically create a type that has the resource keys as properties, to enable dotting into
        the resources from a context.

        For example, given:

        resources = {'foo': <some resource>, 'bar': <some other resource>}

        then this will create the type Resource(namedtuple('foo bar'))

        and then binds the specified resources into an instance of this object, which can be consumed
        as, e.g., context.resources.foo.
        '''
        resource_instance_dict = (
            mapper_fn(self.resource_instance_dict, required_resource_keys)
            if (mapper_fn and required_resource_keys)
            else self.resource_instance_dict
        )

        resource_type = namedtuple('Resources', list(resource_instance_dict.keys()))
        return resource_type(**resource_instance_dict)



import contextlib
import errno
import inspect
import multiprocessing
import os
import re
import subprocess
import tempfile

import yaml

from dagster import check
from dagster.seven.abc import Mapping

from .yaml_utils import load_yaml_from_glob_list, load_yaml_from_globs, load_yaml_from_path


PICKLE_PROTOCOL = 2


DEFAULT_REPOSITORY_YAML_FILENAME = 'repository.yaml'


def file_relative_path(dunderfile, relative_path):
    '''
    This function is useful when one needs to load a file that is
    relative to the position of the current file. (Such as when
    you encode a configuration file path in source file and want
    in runnable in any current working directory)

    It is meant to be used like the following:

    file_relative_path(__file__, 'path/relative/to/file')

    '''

    check.str_param(dunderfile, 'dunderfile')
    check.str_param(relative_path, 'relative_path')

    return os.path.join(os.path.dirname(dunderfile), relative_path)


def script_relative_path(file_path):
    '''
    Useful for testing with local files. Use a path relative to where the
    test resides and this function will return the absolute path
    of that file. Otherwise it will be relative to script that
    ran the test

    Note: this is function is very, very expensive (on the order of 1
    millisecond per invocation) so this should only be used in performance
    insensitive contexts. Prefer file_relative_path for anything with
    performance constraints.

    '''
    # from http://bit.ly/2snyC6s

    check.str_param(file_path, 'file_path')
    scriptdir = inspect.stack()[1][1]
    return os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(scriptdir)), file_path))


# Adapted from https://github.com/okunishinishi/python-stringcase/blob/master/stringcase.py
def camelcase(string):
    check.str_param(string, 'string')

    string = re.sub(r'^[\-_\.]', '', str(string))
    if not string:
        return string
    return str(string[0]).upper() + re.sub(
        r'[\-_\.\s]([a-z])', lambda matched: str(matched.group(1)).upper(), string[1:]
    )


def single_item(ddict):
    check.dict_param(ddict, 'ddict')
    check.param_invariant(len(ddict) == 1, 'ddict')
    return list(ddict.items())[0]


@contextlib.contextmanager
def pushd(path):
    old_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(old_cwd)


def safe_isfile(path):
    '''"Backport of Python 3.8 os.path.isfile behavior.

    This is intended to backport https://docs.python.org/dev/whatsnew/3.8.html#os-path. I'm not
    sure that there are other ways to provoke this behavior on Unix other than the null byte,
    but there are certainly other ways to do it on Windows. Afaict, we won't mask other
    ValueErrors, and the behavior in the status quo ante is rough because we risk throwing an
    unexpected, uncaught ValueError from very deep in our logic.
    '''
    try:
        return os.path.isfile(path)
    except ValueError:
        return False


def mkdir_p(path):
    try:
        os.makedirs(path)
        return path
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def merge_dicts(left, right):
    check.dict_param(left, 'left')
    check.dict_param(right, 'right')

    result = left.copy()
    result.update(right)
    return result


class frozendict(dict):
    def __readonly__(self, *args, **kwargs):
        raise RuntimeError("Cannot modify ReadOnlyDict")

    # https://docs.python.org/3/library/pickle.html#object.__reduce__
    #
    # For a dict, the default behavior for pickle is to iteratively call __setitem__ (see 5th item
    #  in __reduce__ tuple). Since we want to disable __setitem__ and still inherit dict, we
    # override this behavior by defining __reduce__. We return the 3rd item in the tuple, which is
    # passed to __setstate__, allowing us to restore the frozendict.

    def __reduce__(self):
        return (frozendict, (), dict(self))

    def __setstate__(self, state):
        self.__init__(state)

    __setitem__ = __readonly__
    __delitem__ = __readonly__
    pop = __readonly__
    popitem = __readonly__
    clear = __readonly__
    update = __readonly__
    setdefault = __readonly__
    del __readonly__


class frozenlist(list):
    def __readonly__(self, *args, **kwargs):
        raise RuntimeError("Cannot modify ReadOnlyList")

    __setitem__ = __readonly__
    __delitem__ = __readonly__
    append = __readonly__
    clear = __readonly__
    extend = __readonly__
    insert = __readonly__
    pop = __readonly__
    remove = __readonly__
    reverse = __readonly__
    sort = __readonly__


def make_readonly_value(value):
    if isinstance(value, list):
        return frozenlist(list(map(make_readonly_value, value)))
    elif isinstance(value, dict):
        return frozendict({key: make_readonly_value(value) for key, value in value.items()})
    else:
        return value


def get_prop_or_key(elem, key):
    if isinstance(elem, Mapping):
        return elem.get(key)
    else:
        return getattr(elem, key)


def list_pull(alist, key):
    return list(map(lambda elem: get_prop_or_key(elem, key), alist))


def get_multiprocessing_context():
    # Set execution method to spawn, to avoid fork and to have same behavior between platforms.
    # Older versions are stuck with whatever is the default on their platform (fork on
    # Unix-like and spawn on windows)
    #
    # https://docs.python.org/3/library/multiprocessing.html#multiprocessing.get_context
    if hasattr(multiprocessing, 'get_context'):
        return multiprocessing.get_context('spawn')
    else:
        return multiprocessing


def all_none(kwargs):
    for value in kwargs.values():
        if value is not None:
            return False
    return True


def check_script(path):
    subprocess.check_output(['python', path])


def check_cli_execute_file_pipeline(path, pipeline_fn_name, env_file=None):
    cli_cmd = ['python', '-m', 'dagster', 'pipeline', 'execute', '-f', path, '-n', pipeline_fn_name]

    if env_file:
        cli_cmd.append('-e')
        cli_cmd.append(env_file)

    try:
        subprocess.check_output(cli_cmd)
    except subprocess.CalledProcessError as cpe:
        print(cpe)
        raise cpe


def safe_tempfile_path():
    # This gets a valid temporary file path in the safest possible way, although there is still no
    # guarantee that another process will not create a file at this path. The NamedTemporaryFile is
    # deleted when the context manager exits and the file object is closed.
    with tempfile.NamedTemporaryFile() as fd:
        path = fd.name

    return path



import sys

if sys.version_info.major >= 3:
    import typing

    class BuiltinEnum:

        ANY = typing.Any
        BOOL = typing.NewType('Bool', bool)
        FLOAT = typing.NewType('Float', float)
        INT = typing.NewType('Int', int)
        PATH = typing.NewType('Path', str)
        STRING = typing.NewType('String', str)
        NOTHING = typing.NewType('Nothing', None)

        @classmethod
        def contains(cls, value):
            return any(value == getattr(cls, key) for key in dir(cls))


else:
    from enum import Enum

    class BuiltinEnum(Enum):

        ANY = 'Any'
        BOOL = 'Bool'
        FLOAT = 'Float'
        INT = 'Int'
        PATH = 'Path'
        STRING = 'String'
        NOTHING = 'Nothing'

        @classmethod
        def contains(cls, value):
            return isinstance(value, cls)



import sys
import os
import click

from ..version import __version__
from .run import create_run_cli_group
from .pipeline import create_pipeline_cli_group


def create_dagster_cli():
    @click.group(commands={'pipeline': create_pipeline_cli_group(), 'run': create_run_cli_group()})
    @click.version_option(version=__version__)
    def group():
        pass

    # add the path for the cwd so imports in dynamically loaded code work correctly
    sys.path.append(os.getcwd())
    return group


def main():
    cli = create_dagster_cli()
    cli(obj={})  # pylint:disable=E1123



import pytest

from dagster import check
from dagster.utils.yaml_utils import (
    load_yaml_from_glob_list,
    load_yaml_from_globs,
    load_yaml_from_path,
)

from dagster.utils import script_relative_path


def test_load_yaml():
    assert load_yaml_from_path(script_relative_path('yamls/yaml_one.yaml')) == {
        'key_one': {'key_one_one': 'value_one'}
    }


def test_from_glob_list():
    assert load_yaml_from_glob_list([script_relative_path('yamls/yaml_one.yaml')]) == {
        'key_one': {'key_one_one': 'value_one'}
    }

    assert load_yaml_from_glob_list(
        [script_relative_path('yamls/yaml_one.yaml'), script_relative_path('yamls/yaml_two.yaml')]
    ) == {'key_one': {'key_one_one': 'value_one', 'key_one_two': 'value_two'}}

    assert load_yaml_from_glob_list([script_relative_path('yamls/*.yaml')]) == {
        'key_one': {'key_one_one': 'value_one', 'key_one_two': 'value_two'}
    }

    assert load_yaml_from_globs(
        script_relative_path('yamls/yaml_one.yaml'), script_relative_path('yamls/yaml_two.yaml')
    ) == {'key_one': {'key_one_one': 'value_one', 'key_one_two': 'value_two'}}

    with pytest.raises(check.CheckError):
        load_yaml_from_glob_list(['flskhfhjsdf'])



from abc import ABCMeta, abstractproperty, abstractmethod

import six

from dagster import check

from .step import StepExecutionContext
from .system import SystemComputeExecutionContext


class AbstractComputeExecutionContext(six.with_metaclass(ABCMeta)):  # pylint: disable=no-init
    @abstractmethod
    def has_tag(self, key):
        pass

    @abstractmethod
    def get_tag(self, key):
        pass

    @abstractproperty
    def run_id(self):
        pass

    @abstractproperty
    def solid_def(self):
        pass

    @abstractproperty
    def solid(self):
        pass

    @abstractproperty
    def pipeline_def(self):
        pass

    @abstractproperty
    def resources(self):
        pass

    @abstractproperty
    def log(self):
        pass


class ComputeExecutionContext(StepExecutionContext, AbstractComputeExecutionContext):
    __slots__ = ['_system_compute_execution_context']

    def __init__(self, system_compute_execution_context):
        self._system_compute_execution_context = check.inst_param(
            system_compute_execution_context,
            'system_compute_execution_context',
            SystemComputeExecutionContext,
        )
        super(ComputeExecutionContext, self).__init__(system_compute_execution_context)

    @property
    def solid_config(self):
        return self._system_compute_execution_context.solid_config



from collections import namedtuple
from dagster import check

from dagster.core.types.runtime import RuntimeType, resolve_to_runtime_type

from .expectation import IOExpectationDefinition
from .utils import check_valid_name


class InputDefinition(object):
    '''An InputDefinition instance represents an argument to a compute function defined within
    a solid. Inputs are values within the dagster type system that are created from previous
    solids.

    Args:
        name (str): Name of the input.
        dagster_type (DagsterType):
            Type of the input. Defaults to :py:class:`Any` . Basic python types will be
            mapped to the appropriate DagsterType.
        expectations (Optional[List[IOExpectationDefinition]]):
            **Deprecated**: List of expectations that applies to the value passed to the solid.

            Prefer yielding :py:class:`ExpectationResult` from solid compute function.
        description (str): Description of the input. Optional.
    '''

    def __init__(self, name, dagster_type=None, expectations=None, description=None):
        ''
        self.name = check_valid_name(name)

        self.runtime_type = check.inst(resolve_to_runtime_type(dagster_type), RuntimeType)

        self.expectations = check.opt_list_param(
            expectations, 'expectations', of_type=IOExpectationDefinition
        )
        self.description = check.opt_str_param(description, 'description')

    @property
    def descriptive_key(self):
        return 'output'

    def mapping_to(self, solid_name, input_name):
        return InputMapping(self, solid_name, input_name)


class InputMapping(namedtuple('_InputMapping', 'definition solid_name input_name')):
    def __new__(cls, definition, solid_name, input_name):
        return super(InputMapping, cls).__new__(
            cls,
            check.inst_param(definition, 'definition', InputDefinition),
            check.str_param(solid_name, 'solid_name'),
            check.str_param(input_name, 'input_name'),
        )



from __future__ import unicode_literals

import subprocess

from dagster.version import __version__


def test_version():
    assert subprocess.check_output(['dagster', '--version']).decode('utf-8').strip('\n').strip(
        '\r'
    ) == 'dagster, version {version}'.format(version=__version__)



from abc import ABCMeta, abstractmethod
import pickle

import six

from dagster import check
from dagster.utils import PICKLE_PROTOCOL


class SerializationStrategy(six.with_metaclass(ABCMeta)):  # pylint: disable=no-init
    '''This is the base class for serialization / deserialization of dagster objects. The primary
    use case is serialize to / from files, but any file-like objects (e.g. byte buffers) are
    supported.
    '''

    @abstractmethod
    def serialize(self, value, write_file_obj):
        pass

    @abstractmethod
    def deserialize(self, read_file_obj):
        pass

    def serialize_to_file(self, value, write_path):
        check.str_param(write_path, 'write_path')

        with open(write_path, 'wb') as write_obj:
            return self.serialize(value, write_obj)

    def deserialize_from_file(self, read_path):
        check.str_param(read_path, 'read_path')

        with open(read_path, 'rb') as read_obj:
            return self.deserialize(read_obj)


class PickleSerializationStrategy(SerializationStrategy):  # pylint: disable=no-init
    def serialize(self, value, write_file_obj):
        pickle.dump(value, write_file_obj, PICKLE_PROTOCOL)

    def deserialize(self, read_file_obj):
        return pickle.load(read_file_obj)



from abc import ABCMeta, abstractmethod
from collections import defaultdict, namedtuple

import six

from dagster import check
from dagster.utils import camelcase
from dagster.core.errors import DagsterInvalidDefinitionError

from .input import InputDefinition
from .output import OutputDefinition
from .utils import DEFAULT_OUTPUT, struct_to_string


class SolidInvocation(namedtuple('Solid', 'name alias resource_mapper_fn')):
    '''
    A solid identifier in a dependency structure. Allows supplying parameters to the solid,
    like the alias.

    Args:
        name (str): Name of the solid in the pipeline to instance.
        alias (Optional[str]):
            Name for this instance of the solid. Necessary when there are multiple instances
            of the same solid.

    Example:

        .. code-block:: python

            pipeline = Pipeline(
                solid_defs=[solid_1, solid_2]
                dependencies={
                    SolidInvocation('solid_2', alias='other_name') : {
                        'input_name' : DependencyDefinition('solid_2'),
                    },
                    'solid_1' : {
                        'input_name': DependencyDefinition('other_name'),
                    },
                }
            )
    '''

    def __new__(cls, name, alias=None, resource_mapper_fn=None):
        name = check.str_param(name, 'name')
        alias = check.opt_str_param(alias, 'alias')
        resource_mapper_fn = check.opt_callable_param(
            resource_mapper_fn, 'resource_mapper_fn', SolidInvocation.default_resource_mapper_fn
        )
        return super(cls, SolidInvocation).__new__(cls, name, alias, resource_mapper_fn)

    @staticmethod
    def default_resource_mapper_fn(resources, resource_deps):
        return {r: resources.get(r) for r in resource_deps}


class Solid(object):
    '''
    Solid invocation within a pipeline. Defined by its name inside the pipeline.

    Attributes:
        name (str):
            Name of the solid inside the pipeline. Must be unique per-pipeline.
        definition (SolidDefinition):
            Definition of the solid.
    '''

    def __init__(self, name, definition, resource_mapper_fn, container_definition=None):
        from .solid import ISolidDefinition, CompositeSolidDefinition

        self.name = check.str_param(name, 'name')
        self.definition = check.inst_param(definition, 'definition', ISolidDefinition)
        self.resource_mapper_fn = check.callable_param(resource_mapper_fn, 'resource_mapper_fn')
        self.container_definition = check.opt_inst_param(
            container_definition, 'container_definition', CompositeSolidDefinition
        )

        input_handles = {}
        for name, input_def in self.definition.input_dict.items():
            input_handles[name] = SolidInputHandle(self, input_def)

        self._input_handles = input_handles

        output_handles = {}
        for name, output_def in self.definition.output_dict.items():
            output_handles[name] = SolidOutputHandle(self, output_def)

        self._output_handles = output_handles

    def input_handles(self):
        return self._input_handles.values()

    def output_handles(self):
        return self._output_handles.values()

    def input_handle(self, name):
        check.str_param(name, 'name')
        return self._input_handles[name]

    def output_handle(self, name):
        check.str_param(name, 'name')
        return self._output_handles[name]

    def has_input(self, name):
        return self.definition.has_input(name)

    def input_def_named(self, name):
        return self.definition.input_def_named(name)

    def has_output(self, name):
        return self.definition.has_output(name)

    def output_def_named(self, name):
        return self.definition.output_def_named(name)

    @property
    def input_dict(self):
        return self.definition.input_dict

    @property
    def output_dict(self):
        return self.definition.output_dict

    @property
    def step_metadata_fn(self):
        return self.definition.step_metadata_fn

    def container_maps_input(self, input_name):
        if self.container_definition is None:
            return False

        return self.container_definition.mapped_input(self.name, input_name) is not None

    def container_mapped_input(self, input_name):
        return self.container_definition.mapped_input(self.name, input_name)


class SolidHandle(namedtuple('_SolidHandle', 'name definition_name parent')):
    def __new__(cls, name, definition_name, parent):
        return super(SolidHandle, cls).__new__(
            cls,
            check.str_param(name, 'name'),
            check.opt_str_param(definition_name, 'definition_name'),
            check.opt_inst_param(parent, 'parent', SolidHandle),
        )

    def __str__(self):
        return self.to_string()

    def to_string(self):
        # Return unique name of the solid and its lineage (omits solid definition names)
        return self.parent.to_string() + '.' + self.name if self.parent else self.name

    def camelcase(self):
        return (
            self.parent.camelcase() + '.' + camelcase(self.name)
            if self.parent
            else camelcase(self.name)
        )

    @classmethod
    def from_dict(cls, dict_repr):
        '''This method makes it possible to rehydrate a potentially nested SolidHandle after a
        roundtrip through json.loads(json.dumps(SolidHandle._asdict()))'''

        check.dict_param(dict_repr, 'dict_repr', key_type=str)
        check.invariant(
            'name' in dict_repr, 'Dict representation of SolidHandle must have a \'name\' key'
        )
        check.invariant(
            'definition_name' in dict_repr,
            'Dict representation of SolidHandle must have a \'definition_name\' key',
        )
        check.invariant(
            'parent' in dict_repr, 'Dict representation of SolidHandle must have a \'parent\' key'
        )

        if isinstance(dict_repr['parent'], (list, tuple)):
            dict_repr['parent'] = SolidHandle.from_dict(
                {
                    'name': dict_repr['parent'][0],
                    'definition_name': dict_repr['parent'][1],
                    'parent': dict_repr['parent'][2],
                }
            )

        return SolidHandle(**dict_repr)


class SolidInputHandle(namedtuple('_SolidInputHandle', 'solid input_def')):
    def __new__(cls, solid, input_def):
        return super(SolidInputHandle, cls).__new__(
            cls,
            check.inst_param(solid, 'solid', Solid),
            check.inst_param(input_def, 'input_def', InputDefinition),
        )

    def _inner_str(self):
        return struct_to_string(
            'SolidInputHandle',
            solid_name=self.solid.name,
            definition_name=self.solid.definition.name,
            input_name=self.input_def.name,
        )

    def __str__(self):
        return self._inner_str()

    def __repr__(self):
        return self._inner_str()

    def __hash__(self):
        return hash((self.solid.name, self.input_def.name))

    def __eq__(self, other):
        return self.solid.name == other.solid.name and self.input_def.name == other.input_def.name


class SolidOutputHandle(namedtuple('_SolidOutputHandle', 'solid output_def')):
    def __new__(cls, solid, output_def):
        return super(SolidOutputHandle, cls).__new__(
            cls,
            check.inst_param(solid, 'solid', Solid),
            check.inst_param(output_def, 'output_def', OutputDefinition),
        )

    def _inner_str(self):
        return struct_to_string(
            'SolidOutputHandle',
            solid_name=self.solid.name,
            definition_name=self.solid.definition.name,
            output_name=self.output_def.name,
        )

    def __str__(self):
        return self._inner_str()

    def __repr__(self):
        return self._inner_str()

    def __hash__(self):
        return hash((self.solid.name, self.output_def.name))

    def __eq__(self, other):
        return self.solid.name == other.solid.name and self.output_def.name == other.output_def.name


class InputToOutputHandleDict(defaultdict):
    def __init__(self):
        defaultdict.__init__(self, list)

    def __getitem__(self, key):
        check.inst_param(key, 'key', SolidInputHandle)
        return defaultdict.__getitem__(self, key)

    def __setitem__(self, key, val):
        check.inst_param(key, 'key', SolidInputHandle)
        if not (isinstance(val, SolidOutputHandle) or isinstance(val, list)):
            check.failed(
                'Value must be SolidOutoutHandle or List[SolidOutputHandle], got {val}'.format(
                    val=type(val)
                )
            )

        return defaultdict.__setitem__(self, key, val)


def _create_handle_dict(solid_dict, dep_dict):
    check.dict_param(solid_dict, 'solid_dict', key_type=str, value_type=Solid)
    check.two_dim_dict_param(dep_dict, 'dep_dict', value_type=IDependencyDefinition)

    handle_dict = InputToOutputHandleDict()

    for solid_name, input_dict in dep_dict.items():
        from_solid = solid_dict[solid_name]
        for input_name, dep_def in input_dict.items():
            if dep_def.is_multi():
                handle_dict[from_solid.input_handle(input_name)] = [
                    solid_dict[dep.solid].output_handle(dep.output)
                    for dep in dep_def.get_definitions()
                ]
            else:
                handle_dict[from_solid.input_handle(input_name)] = solid_dict[
                    dep_def.solid
                ].output_handle(dep_def.output)

    return handle_dict


class DependencyStructure(object):
    @staticmethod
    def from_definitions(solids, dep_dict):
        return DependencyStructure(_create_handle_dict(solids, dep_dict))

    def __init__(self, handle_dict):
        self._handle_dict = check.inst_param(handle_dict, 'handle_dict', InputToOutputHandleDict)

    def deps_of_solid(self, solid_name):
        check.str_param(solid_name, 'solid_name')

        return list(handles[1] for handles in self.__gen_deps_of_solid(solid_name))

    def deps_of_solid_with_input(self, solid_name):
        check.str_param(solid_name, 'solid_name')
        result = defaultdict(list)
        for input_handle, output_handle in self.__gen_deps_of_solid(solid_name):
            result[input_handle].append(output_handle)
        return result

    def __gen_deps_of_solid(self, solid_name):
        for input_handle, output_handle_or_list in self._handle_dict.items():
            if input_handle.solid.name == solid_name:
                if isinstance(output_handle_or_list, list):
                    for output_handle in output_handle_or_list:
                        yield (input_handle, output_handle)
                else:
                    yield (input_handle, output_handle_or_list)

    def depended_by_of_solid(self, solid_name):
        check.str_param(solid_name, 'solid_name')
        result = defaultdict(list)
        for input_handle, output_handle_or_list in self._handle_dict.items():
            if isinstance(output_handle_or_list, list):
                for output_handle in output_handle_or_list:
                    if output_handle.solid.name == solid_name:
                        result[output_handle].append(input_handle)
            else:
                if output_handle_or_list.solid.name == solid_name:
                    result[output_handle_or_list].append(input_handle)

        return result

    def has_singular_dep(self, solid_input_handle):
        check.inst_param(solid_input_handle, 'solid_input_handle', SolidInputHandle)
        return isinstance(self._handle_dict.get(solid_input_handle), SolidOutputHandle)

    def get_singular_dep(self, solid_input_handle):
        check.inst_param(solid_input_handle, 'solid_input_handle', SolidInputHandle)
        dep = self._handle_dict[solid_input_handle]
        check.invariant(
            isinstance(dep, SolidOutputHandle),
            'Can not call get_singular_dep when dep is not singular, got {dep}'.format(
                dep=type(dep)
            ),
        )
        return dep

    def has_multi_deps(self, solid_input_handle):
        check.inst_param(solid_input_handle, 'solid_input_handle', SolidInputHandle)
        return isinstance(self._handle_dict.get(solid_input_handle), list)

    def get_multi_deps(self, solid_input_handle):
        check.inst_param(solid_input_handle, 'solid_input_handle', SolidInputHandle)
        dep = self._handle_dict[solid_input_handle]
        check.invariant(
            isinstance(dep, list),
            'Can not call get_multi_dep when dep is singular, got {dep}'.format(dep=type(dep)),
        )
        return dep

    def has_deps(self, solid_input_handle):
        check.inst_param(solid_input_handle, 'solid_input_handle', SolidInputHandle)
        return solid_input_handle in self._handle_dict

    def get_deps_list(self, solid_input_handle):
        check.inst_param(solid_input_handle, 'solid_input_handle', SolidInputHandle)
        check.invariant(self.has_deps(solid_input_handle))
        if self.has_singular_dep(solid_input_handle):
            return [self.get_singular_dep(solid_input_handle)]
        else:
            return self.get_multi_deps(solid_input_handle)

    def input_handles(self):
        return list(self._handle_dict.keys())

    def items(self):
        return self._handle_dict.items()


class IDependencyDefinition(six.with_metaclass(ABCMeta)):  # pylint: disable=no-init
    @abstractmethod
    def get_definitions(self):
        pass

    @abstractmethod
    def is_multi(self):
        pass


class DependencyDefinition(
    namedtuple('_DependencyDefinition', 'solid output description'), IDependencyDefinition
):
    '''Dependency definitions represent an edge in the DAG of solids. This object is
    used with a dictionary structure (whose keys represent solid/input where the dependency
    comes from) so this object only contains the target dependency information.

    Args:
        solid (str):
            The name of the solid that is the target of the dependency.
            This is the solid where the value passed between the solids
            comes from.
        output (str):
            The name of the output that is the target of the dependency.
            Defaults to "result", the default output name of solids with a single output.
        description (str):
            Description of this dependency. Optional.
    '''

    def __new__(cls, solid, output=DEFAULT_OUTPUT, description=None):
        return super(DependencyDefinition, cls).__new__(
            cls,
            check.str_param(solid, 'solid'),
            check.str_param(output, 'output'),
            check.opt_str_param(description, 'description'),
        )

    def get_definitions(self):
        return [self]

    def is_multi(self):
        return False


class MultiDependencyDefinition(
    namedtuple('_MultiDependencyDefinition', 'dependencies'), IDependencyDefinition
):
    def __new__(cls, dependencies):
        deps = check.list_param(dependencies, 'dependencies', of_type=DependencyDefinition)
        seen = {}
        for dep in deps:
            key = dep.solid + ':' + dep.output
            if key in seen:
                raise DagsterInvalidDefinitionError(
                    'Duplicate dependencies on solid "{dep.solid}" output "{dep.output}" '
                    'used in the same MultiDependencyDefinition.'.format(dep=dep)
                )
            seen[key] = True

        return super(MultiDependencyDefinition, cls).__new__(cls, deps)

    def get_definitions(self):
        return self.dependencies

    def is_multi(self):
        return True



from dagster import PipelineDefinition, check
from dagster.core.definitions import create_environment_type

from dagster.core.types.config import ConfigType


def scaffold_pipeline_config(pipeline_def, skip_optional=True, mode=None):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.bool_param(skip_optional, 'skip_optional')

    env_config_type = create_environment_type(pipeline_def, mode=mode)

    env_dict = {}

    for env_field_name, env_field in env_config_type.fields.items():
        if skip_optional and env_field.is_optional:
            continue

        # unfortunately we have to treat this special for now
        if env_field_name == 'context':
            if skip_optional and env_config_type.fields['context'].is_optional:
                continue

        env_dict[env_field_name] = scaffold_type(env_field.config_type, skip_optional)

    return env_dict


def scaffold_type(config_type, skip_optional=True):
    check.inst_param(config_type, 'config_type', ConfigType)
    check.bool_param(skip_optional, 'skip_optional')

    # Right now selectors and composites have the same
    # scaffolding logic, which might not be wise.
    if config_type.has_fields:
        default_dict = {}
        for field_name, field in config_type.fields.items():
            if skip_optional and field.is_optional:
                continue

            default_dict[field_name] = scaffold_type(field.config_type, skip_optional)
        return default_dict
    elif config_type.is_any:
        return 'AnyType'
    elif config_type.is_scalar:
        defaults = {'String': '', 'Path': 'path/to/something', 'Int': 0, 'Bool': True}

        return defaults[config_type.name]
    elif config_type.is_list:
        return []
    elif config_type.is_enum:
        return '|'.join(sorted(map(lambda v: v.config_value, config_type.enum_values)))
    else:
        check.failed('Do not know how to scaffold {type_name}'.format(type_name=config_type.name))



from dagster import check

from .runtime import RuntimeType, define_python_dagster_type


def _decorate_as_dagster_type(
    bare_cls,
    name,
    description,
    input_hydration_config=None,
    output_materialization_config=None,
    serialization_strategy=None,
    auto_plugins=None,
    typecheck_metadata_fn=None,
):
    dagster_type_cls = define_python_dagster_type(
        name=name,
        description=description,
        python_type=bare_cls,
        input_hydration_config=input_hydration_config,
        output_materialization_config=output_materialization_config,
        serialization_strategy=serialization_strategy,
        auto_plugins=auto_plugins,
        typecheck_metadata_fn=typecheck_metadata_fn,
    )

    type_inst = dagster_type_cls.inst()

    make_klass_runtime_type_decorated_klass(bare_cls, type_inst)
    return bare_cls


def dagster_type(
    name=None,
    description=None,
    input_hydration_config=None,
    output_materialization_config=None,
    serialization_strategy=None,
    auto_plugins=None,
    typecheck_metadata_fn=None,
):
    '''
    Decorator version of as_dagster_type. See documentation for :py:func:`as_dagster_type` .

    See documentation for :py:func:`define_python_dagster_type` for parameters.

    This allows for the straightforward creation of your own classes for use in your
    business logic, and then annotating them to make those same classes compatible with
    the dagster type system.

    e.g.:

    # You have created an object for your own purposes within your app
    class MyDataObject:
        pass


    Now you want to be able to mark this as an input or output to solid. Without
    modification, this does not work.

    You must decorate it:

    @dagster_type
    class MyDataObject:
        pass

    Now one can using this as an input or an output into a solid.

    @lambda_solid
    def create_myobject() -> MyDataObject:
        return MyDataObject()

    And it is viewable in dagit and so forth, and you can use the dagster type system
    for configuration, serialization, and metadata emission.

    Decorator version of as_dagster_type. See :py:func:`as_dagster_type` for parameter
    documentation. 
    '''

    def _with_args(bare_cls):
        check.type_param(bare_cls, 'bare_cls')
        new_name = name if name else bare_cls.__name__
        return _decorate_as_dagster_type(
            bare_cls=bare_cls,
            name=new_name,
            description=description,
            input_hydration_config=input_hydration_config,
            output_materialization_config=output_materialization_config,
            serialization_strategy=serialization_strategy,
            auto_plugins=auto_plugins,
            typecheck_metadata_fn=typecheck_metadata_fn,
        )

    # check for no args, no parens case
    if callable(name):
        klass = name  # with no parens, name is actually the decorated class
        return _decorate_as_dagster_type(bare_cls=klass, name=klass.__name__, description=None)

    return _with_args


MAGIC_RUNTIME_TYPE_NAME = '__runtime_type'


def is_runtime_type_decorated_klass(klass):
    check.type_param(klass, 'klass')
    return hasattr(klass, MAGIC_RUNTIME_TYPE_NAME)


def get_runtime_type_on_decorated_klass(klass):
    check.type_param(klass, 'klass')
    return getattr(klass, MAGIC_RUNTIME_TYPE_NAME)


def make_klass_runtime_type_decorated_klass(klass, runtime_type):
    check.type_param(klass, 'klass')
    check.inst_param(runtime_type, 'runtime_type', RuntimeType)
    setattr(klass, MAGIC_RUNTIME_TYPE_NAME, runtime_type)


def as_dagster_type(
    existing_type,
    name=None,
    description=None,
    input_hydration_config=None,
    output_materialization_config=None,
    serialization_strategy=None,
    auto_plugins=None,
    typecheck_metadata_fn=None,
):
    '''
    See documentation for :py:func:`define_python_dagster_type` for parameters.

    Takes a python cls and creates a type for it in the Dagster domain.

    Frequently you want to import a data processing library and use its types
    directly in solid definitions. To support this dagster has this facility
    that allows one to annotate *existing* classes as dagster type. 

    Note: It does this by setting a magical property (current "__runtime_type") on the
    class itself pointing to the dagster type associated with the python class

    e.g.

    from existing_library import FancyDataType as ExistingFancyDataType

    FancyDataType = as_dagster_type(existing_type=ExistingFancyDataType, name='FancyDataType')

    While one *could* use the existing type directly from the original library, we would
    recommend using the object retrned by as_dagster_type to avoid an import-order-based bugs.

    See dagster_pandas for an example of how to do this.
    '''

    return _decorate_as_dagster_type(
        bare_cls=check.type_param(existing_type, 'existing_type'),
        name=check.opt_str_param(name, 'name', existing_type.__name__),
        description=description,
        input_hydration_config=input_hydration_config,
        output_materialization_config=output_materialization_config,
        serialization_strategy=serialization_strategy,
        auto_plugins=auto_plugins,
        typecheck_metadata_fn=typecheck_metadata_fn,
    )



import tempfile
import os

import pytest

from dagster.core.types.marshal import PickleSerializationStrategy


# https://dev.azure.com/elementl/dagster/_build/results?buildId=2941
@pytest.mark.skipif(
    os.name == 'nt', reason='Azure pipelines does not let us use tempfile.NamedTemporaryFile'
)
def test_serialization_strategy():
    serialization_strategy = PickleSerializationStrategy()
    with tempfile.NamedTemporaryFile() as fd:
        serialization_strategy.serialize_to_file('foo', fd.name)
        assert serialization_strategy.deserialize_from_file(fd.name) == 'foo'



import dagster

from dagster import (
    DependencyDefinition,
    Dict,
    Field,
    InputDefinition,
    OutputDefinition,
    PipelineDefinition,
    Output,
    SolidDefinition,
    String,
    check,
    execute_pipeline,
)

from dagster.core.test_utils import single_output_transform

SingleValueDict = Dict({'value': Field(String)})


def define_pass_value_solid(name, description=None):
    check.str_param(name, 'name')
    check.opt_str_param(description, 'description')

    def _value_t_fn(context, _inputs):
        yield Output(context.solid_config['value'])

    return SolidDefinition(
        name=name,
        description=description,
        input_defs=[],
        output_defs=[OutputDefinition(String)],
        config_field=Field(SingleValueDict),
        compute_fn=_value_t_fn,
    )


def test_execute_solid_with_input_same_name():
    a_thing_solid = single_output_transform(
        'a_thing',
        input_defs=[InputDefinition(name='a_thing')],
        compute_fn=lambda context, inputs: inputs['a_thing'] + inputs['a_thing'],
        output_def=dagster.OutputDefinition(),
    )

    pipeline = PipelineDefinition(
        solid_defs=[define_pass_value_solid('pass_value'), a_thing_solid],
        dependencies={'a_thing': {'a_thing': DependencyDefinition('pass_value')}},
    )

    result = execute_pipeline(
        pipeline, environment_dict={'solids': {'pass_value': {'config': {'value': 'foo'}}}}
    )

    assert result.result_for_solid('a_thing').output_value() == 'foofoo'


def test_execute_two_solids_with_same_input_name():
    input_def = InputDefinition(name='a_thing')

    solid_one = single_output_transform(
        'solid_one',
        input_defs=[input_def],
        compute_fn=lambda context, inputs: inputs['a_thing'] + inputs['a_thing'],
        output_def=dagster.OutputDefinition(),
    )

    solid_two = single_output_transform(
        'solid_two',
        input_defs=[input_def],
        compute_fn=lambda context, inputs: inputs['a_thing'] + inputs['a_thing'],
        output_def=dagster.OutputDefinition(),
    )

    pipeline = dagster.PipelineDefinition(
        solid_defs=[
            define_pass_value_solid('pass_to_one'),
            define_pass_value_solid('pass_to_two'),
            solid_one,
            solid_two,
        ],
        dependencies={
            'solid_one': {'a_thing': DependencyDefinition('pass_to_one')},
            'solid_two': {'a_thing': DependencyDefinition('pass_to_two')},
        },
    )

    result = execute_pipeline(
        pipeline,
        environment_dict={
            'solids': {
                'pass_to_one': {'config': {'value': 'foo'}},
                'pass_to_two': {'config': {'value': 'bar'}},
            }
        },
    )

    assert result.success

    assert result.result_for_solid('solid_one').output_value() == 'foofoo'
    assert result.result_for_solid('solid_two').output_value() == 'barbar'


def test_execute_dep_solid_different_input_name():
    pass_to_first = define_pass_value_solid('pass_to_first')

    first_solid = single_output_transform(
        'first_solid',
        input_defs=[InputDefinition(name='a_thing')],
        compute_fn=lambda context, inputs: inputs['a_thing'] + inputs['a_thing'],
        output_def=dagster.OutputDefinition(),
    )

    second_solid = single_output_transform(
        'second_solid',
        input_defs=[InputDefinition(name='an_input')],
        compute_fn=lambda context, inputs: inputs['an_input'] + inputs['an_input'],
        output_def=dagster.OutputDefinition(),
    )

    pipeline = dagster.PipelineDefinition(
        solid_defs=[pass_to_first, first_solid, second_solid],
        dependencies={
            'first_solid': {'a_thing': DependencyDefinition('pass_to_first')},
            'second_solid': {'an_input': DependencyDefinition('first_solid')},
        },
    )

    result = dagster.execute_pipeline(
        pipeline, environment_dict={'solids': {'pass_to_first': {'config': {'value': 'bar'}}}}
    )

    assert result.success
    assert len(result.solid_result_list) == 3
    assert result.result_for_solid('pass_to_first').output_value() == 'bar'
    assert result.result_for_solid('first_solid').output_value() == 'barbar'
    assert result.result_for_solid('second_solid').output_value() == 'barbarbarbar'



from dagster.utils import camelcase


def test_camelcase():
    assert camelcase('foo') == 'Foo'
    assert camelcase('foo_bar') == 'FooBar'
    assert camelcase('foo.bar') == 'FooBar'
    assert camelcase('foo-bar') == 'FooBar'
    assert camelcase('') == ''



import datetime
import itertools
import logging
import uuid

from collections import OrderedDict, namedtuple

from dagster import check, seven
from dagster.utils import frozendict

DAGSTER_META_KEY = 'dagster_meta'


PYTHON_LOGGING_LEVELS_MAPPING = frozendict(
    OrderedDict({'CRITICAL': 50, 'ERROR': 40, 'WARNING': 30, 'INFO': 20, 'DEBUG': 10})
)

PYTHON_LOGGING_LEVELS_ALIASES = frozendict(OrderedDict({'FATAL': 'CRITICAL', 'WARN': 'WARNING'}))

PYTHON_LOGGING_LEVELS_NAMES = frozenset(
    [
        level_name.lower()
        for level_name in sorted(
            list(PYTHON_LOGGING_LEVELS_MAPPING.keys()) + list(PYTHON_LOGGING_LEVELS_ALIASES.keys())
        )
    ]
)


def _dump_value(value):
    # dump namedtuples as objects instead of arrays
    if isinstance(value, tuple) and hasattr(value, '_asdict'):
        return seven.json.dumps(value._asdict())

    return seven.json.dumps(value)


def _kv_message(all_items):
    sep = '\n'
    format_str = '{key:>20} = {value}'
    return sep + sep.join(
        [format_str.format(key=key, value=_dump_value(value)) for key, value in all_items]
    )


def coerce_valid_log_level(log_level):
    '''Convert a log level into an integer for consumption by the low-level Python logging API.'''
    if isinstance(log_level, int):
        return log_level
    check.str_param(log_level, 'log_level')
    check.invariant(
        log_level.lower() in PYTHON_LOGGING_LEVELS_NAMES,
        'Bad value for log level {level}: permissible values are {levels}.'.format(
            level=log_level,
            levels=', '.join(['\'{level_name}\'' for level_name in PYTHON_LOGGING_LEVELS_NAMES]),
        ),
    )
    log_level = PYTHON_LOGGING_LEVELS_ALIASES.get(log_level.upper(), log_level.upper())
    return PYTHON_LOGGING_LEVELS_MAPPING[log_level]


class DagsterLogManager(namedtuple('_DagsterLogManager', 'run_id logging_tags loggers')):
    '''Centralized dispatch for logging through the execution context.

    Handles the construction of uniform structured log messages and passes through to the underlying
    loggers.

    An instance of the log manager is made available to solids as context.log.

    The log manager supports standard convenience methods like those built into the Python logging
    library (i.e., ``DagsterLogManager.{debug, info, warning, error, critical}``, expects
    corresponding methods to be defined on the loggers passed to its constructor, and will delegate
    to those methods on each logger.

    The underlying integer API can also be called directly using, e.g.,
    ``DagsterLogManager.log(5, msg)``, and the log manager will delegate to the `log` method defined
    on each of the loggers it manages.

    User-defined custom log levels are not supported, and calls to, e.g.,
    ``DagsterLogManager.trace()`` or ``DagsterLogManager.notice`` will result in hard exceptions at
    runtime.

    Args:
        run_id (str): The run_id.
        tags (dict): Tags for the run
        loggers (List[logging.Logger]): Loggers to invoke.
    '''

    def __new__(cls, run_id, logging_tags, loggers):
        return super(DagsterLogManager, cls).__new__(
            cls,
            run_id=check.str_param(run_id, 'run_id'),
            logging_tags=check.dict_param(logging_tags, 'logging_tags'),
            loggers=check.list_param(loggers, 'loggers', of_type=logging.Logger),
        )

    def _prepare_message(self, orig_message, message_props):
        check.str_param(orig_message, 'orig_message')
        check.dict_param(message_props, 'message_props')

        # These are todos to further align with the Python logging API
        check.invariant(
            'extra' not in message_props, 'do not allow until explicit support is handled'
        )
        check.invariant(
            'exc_info' not in message_props, 'do not allow until explicit support is handled'
        )

        # Reserved keys in the message_props -- these are system generated.
        check.invariant('orig_message' not in message_props, 'orig_message reserved value')
        check.invariant('message' not in message_props, 'message reserved value')
        check.invariant('log_message_id' not in message_props, 'log_message_id reserved value')
        check.invariant('log_timestamp' not in message_props, 'log_timestamp reserved value')

        log_message_id = str(uuid.uuid4())

        log_timestamp = datetime.datetime.utcnow().isoformat()

        synth_props = {
            'orig_message': orig_message,
            'log_message_id': log_message_id,
            'log_timestamp': log_timestamp,
            'run_id': self.run_id,
        }

        # We first generate all props for the purpose of producing the semi-structured
        # log message via _kv_messsage
        all_props = dict(
            itertools.chain(synth_props.items(), self.logging_tags.items(), message_props.items())
        )

        # So here we use the arbitrary key DAGSTER_META_KEY to store a dictionary of
        # all the meta information that dagster injects into log message.
        # The python logging module, in its infinite wisdom, actually takes all the
        # keys in extra and unconditionally smashes them into the internal dictionary
        # of the logging.LogRecord class. We used a reserved key here to avoid naming
        # collisions with internal variables of the LogRecord class.
        # See __init__.py:363 (makeLogRecord) in the python 3.6 logging module source
        # for the gory details.
        return (_kv_message(all_props.items()), {DAGSTER_META_KEY: all_props})

    def _log(self, level, orig_message, message_props):
        '''Actually invoke the underlying loggers for a given log level.

        Args:
            level (Union[str, int]): An integer represeting a Python logging level or one of the
                standard Python string representations of a loggging level.
            orig_message (str): The log message generated in user code.
            message_props (dict): Additional properties for the structured log message.
        '''
        if not self.loggers:
            return

        level = coerce_valid_log_level(level)

        message, extra = self._prepare_message(orig_message, message_props)

        for logger_ in self.loggers:
            logger_.log(level, message, extra=extra)

    def debug(self, msg, **kwargs):
        '''
        Debug level logging directive. Ends up invoking loggers with DEBUG error level.

        The message will be automatically adorned with context information about the name
        of the pipeline, the name of the solid, and so forth. The use can also add
        context values during execution using the value() method of ExecutionContext.
        Therefore it is generally unnecessary to include this type of information
        (solid name, pipeline name, etc) in the log message unless it is critical
        for the readability/fluency of the log message text itself.

        You can optionally additional context key-value pairs to an individual log
        message using the keyword args to this message

        Args:
            msg (str): The message to log.

        Kwargs:
            Additional context values for only this log message.
        '''

        check.str_param(msg, 'msg')
        return self._log(logging.DEBUG, msg, kwargs)

    def info(self, msg, **kwargs):
        '''Log at INFO level

        See ``debug()``.

        Args:
            msg (str): The message to log.

        Kwargs:
            Additional context values for only this log message.
        '''

        check.str_param(msg, 'msg')
        return self._log(logging.INFO, msg, kwargs)

    def warning(self, msg, **kwargs):
        '''Log at WARNING level

        See ``debug()``.

        Args:
            msg (str): The message to log.

        Kwargs:
            Additional context values for only this log message.
        '''

        check.str_param(msg, 'msg')
        return self._log(logging.WARNING, msg, kwargs)

    # Define the alias .warn()
    warn = warning

    def error(self, msg, **kwargs):
        '''Log at ERROR level

        See ``debug()``.

        Args:
            msg (str): The message to log.

        Kwargs:
            Additional context values for only this log message.
        '''

        check.str_param(msg, 'msg')
        return self._log(logging.ERROR, msg, kwargs)

    def critical(self, msg, **kwargs):
        '''Log at CRITICAL level

        See ``debug()``.

        Args:
            msg (str): The message to log.

        Kwargs:
            Additional context values for only this log message.
        '''

        check.str_param(msg, 'msg')
        return self._log(logging.CRITICAL, msg, kwargs)

    # Define the alias .fatal()
    fatal = critical



from dagster import check
from dagster.core.errors import DagsterError

from .utils import check_valid_name


class DagsterIOExpectationFailedError(DagsterError):
    '''Thrown with pipeline configured to throw on expectation failure'''

    def __init__(self, expectation_context, value, *args, **kwargs):
        super(DagsterIOExpectationFailedError, self).__init__(*args, **kwargs)
        self.expectation_context = expectation_context
        self.value = value

    def __repr__(self):
        inout_def = self.expectation_context.inout_def
        return (
            'DagsterIOExpectationFailedError(solid={solid_name}, {key}={inout_name}, '
            'expectation={e_name}, value={value})'
        ).format(
            solid_name=self.expectation_context.solid.name,
            key=inout_def.descriptive_key,
            inout_name=inout_def.name,
            e_name=self.expectation_context.expectation_def.name,
            value=repr(self.value),
        )

    def __str__(self):
        return self.__repr__()


class IOExpectationDefinition(object):
    '''
    Expectations represent a data quality test. It performs an arbitrary computation
    to see if a given input or output satisfies the expectation.

    Attributes:

        name (str): The name of the expectation. Names should be unique per-solid.
        expectation_fn (callable):
            This is the implementation of an expectation computation. It should be a callback
            with the signature (**context**: `ExecutionContext`, **info**:
            `ExpectationExecutionInfo`, **value**: `Any`) : `ExpectationResult`.

            "value" conforms to the type check performed within the Dagster type system.

            e.g. If the expectation is declared on an input of type ``dagster_pandas.DataFrame``,
            you can assume that value is a ``pandas.DataFrame``.

        description (str): Description of expectation. Optional.

    Examples:

        .. code-block:: python

            InputDefinition('some_input', types.Int, expectations=[
                IOExpectationDefinition(
                    name='is_positive',
                    expectation_fn=lambda(
                        _info,
                        value,
                    ): ExpectationResult(success=value > 0),
                )
            ])
    '''

    def __init__(self, name, expectation_fn, description=None):
        self.name = check_valid_name(name)
        self.expectation_fn = check.callable_param(expectation_fn, 'expectation_fn')
        self.description = check.opt_str_param(description, 'description')



from collections import namedtuple

from dagster.utils import check
from dagster.core.definitions import SolidHandle
from dagster.core.errors import DagsterInvalidDefinitionError
from dagster.core.types import Bool, Field, List, NamedDict, NamedSelector
from dagster.core.types.config import ALL_CONFIG_BUILTINS, ConfigType, ConfigTypeAttributes
from dagster.core.types.field_utils import FieldImpl, check_opt_field_param, _ConfigComposite
from dagster.core.types.iterate_types import iterate_config_types
from dagster.core.types.runtime import construct_runtime_type_dictionary
from dagster.utils import camelcase, single_item

from .dependency import DependencyStructure, Solid, SolidHandle, SolidInputHandle
from .logger import LoggerDefinition
from .mode import ModeDefinition
from .resource import ResourceDefinition
from .solid import CompositeSolidDefinition, ISolidDefinition, SolidDefinition


def SystemNamedDict(name, fields, description=None):
    '''A SystemNamedDict object is simply a NamedDict intended for internal (dagster) use.
    '''
    return NamedDict(name, fields, description, ConfigTypeAttributes(is_system_config=True))


class _SolidContainerConfigDict(_ConfigComposite):
    def __init__(self, name, fields, description=None, handle=None, child_solids_config_field=None):
        self._handle = check.opt_inst_param(handle, 'handle', SolidHandle)
        self._child_solids_config_field = check.opt_inst_param(
            child_solids_config_field, 'child_solids_config_field', FieldImpl
        )

        super(_SolidContainerConfigDict, self).__init__(
            key=name,
            name=name,
            fields=fields,
            description=description,
            type_attributes=ConfigTypeAttributes(is_system_config=True),
        )

    @property
    def handle(self):
        '''A solid handle ref to the composite solid that is associated with this config schema
        (e.g., this is the top-level config object for a composite solid or pipeline)
        '''
        return self._handle

    @property
    def child_solids_config_field(self):
        '''We stash the config schema for the children of composite solids here so that we can
        continue schema traversal even when masked at the top level of config
        '''
        return self._child_solids_config_field


def SolidContainerConfigDict(
    name, fields, description=None, handle=None, child_solids_config_field=None
):
    class _SolidContainerConfigDictInternal(_SolidContainerConfigDict):
        def __init__(self):
            super(_SolidContainerConfigDictInternal, self).__init__(
                name=name,
                fields=fields,
                description=description,
                handle=handle,
                child_solids_config_field=child_solids_config_field,
            )

    return _SolidContainerConfigDictInternal


def SystemNamedSelector(name, fields, description=None):
    return NamedSelector(name, fields, description, ConfigTypeAttributes(is_system_config=True))


def is_solid_container_config(obj):
    return isinstance(obj, _SolidContainerConfigDict)


def _is_selector_field_optional(config_type):
    check.inst_param(config_type, 'config_type', ConfigType)
    if len(config_type.fields) > 1:
        return False
    else:
        _name, field = single_item(config_type.fields)
        return field.is_optional


def define_resource_cls(parent_name, resource_name, resource_def):
    return SystemNamedDict(
        '{parent_name}.{resource_name}'.format(
            parent_name=parent_name, resource_name=camelcase(resource_name)
        ),
        {'config': resource_def.config_field},
    )


def define_resource_dictionary_cls(name, resource_defs):
    check.str_param(name, 'name')
    check.dict_param(resource_defs, 'resource_defs', key_type=str, value_type=ResourceDefinition)

    fields = {}
    for resource_name, resource_def in resource_defs.items():
        if resource_def.config_field:
            fields[resource_name] = Field(define_resource_cls(name, resource_name, resource_def))

    return SystemNamedDict(name=name, fields=fields)


def remove_none_entries(ddict):
    return {k: v for k, v in ddict.items() if v is not None}


def define_solid_config_cls(name, config_field, inputs_field, outputs_field):
    check.str_param(name, 'name')
    check_opt_field_param(config_field, 'config_field')
    check_opt_field_param(inputs_field, 'inputs_field')
    check_opt_field_param(outputs_field, 'outputs_field')

    return SystemNamedDict(
        name,
        remove_none_entries(
            {'config': config_field, 'inputs': inputs_field, 'outputs': outputs_field}
        ),
    )


class EnvironmentClassCreationData(
    namedtuple(
        'EnvironmentClassCreationData',
        'pipeline_name solids dependency_structure mode_definition logger_defs',
    )
):
    def __new__(cls, pipeline_name, solids, dependency_structure, mode_definition, logger_defs):
        return super(EnvironmentClassCreationData, cls).__new__(
            cls,
            pipeline_name=check.str_param(pipeline_name, 'pipeline_name'),
            solids=check.list_param(solids, 'solids', of_type=Solid),
            dependency_structure=check.inst_param(
                dependency_structure, 'dependency_structure', DependencyStructure
            ),
            mode_definition=check.opt_inst_param(
                mode_definition, 'mode_definition', ModeDefinition
            ),
            logger_defs=check.dict_param(
                logger_defs, 'logger_defs', key_type=str, value_type=LoggerDefinition
            ),
        )


def define_mode_resources_dictionary_cls(pipeline_name, mode_definition):
    check.str_param(pipeline_name, 'pipeline_name')
    check.inst_param(mode_definition, 'mode_definition', ModeDefinition)

    return define_resource_dictionary_cls(
        '{pipeline_name}.Mode.{mode}.Resources'.format(
            pipeline_name=pipeline_name, mode=camelcase(mode_definition.name)
        ),
        mode_definition.resource_defs,
    )


def define_logger_dictionary_cls(name, creation_data):
    check.str_param(name, 'name')
    check.inst_param(creation_data, 'creation_data', EnvironmentClassCreationData)

    fields = {}

    for logger_name, logger_definition in creation_data.logger_defs.items():
        fields[logger_name] = Field(
            SystemNamedDict(
                '{pipeline_name}.LoggerConfig.{logger_name}'.format(
                    pipeline_name=camelcase(creation_data.pipeline_name),
                    logger_name=camelcase(logger_name),
                ),
                remove_none_entries({'config': logger_definition.config_field}),
            ),
            is_optional=True,
        )

    return SystemNamedDict(name, fields)


def define_environment_cls(creation_data):
    check.inst_param(creation_data, 'creation_data', EnvironmentClassCreationData)
    pipeline_name = camelcase(creation_data.pipeline_name)

    return SolidContainerConfigDict(
        name='{pipeline_name}.Mode.{mode_name}.Environment'.format(
            pipeline_name=pipeline_name, mode_name=camelcase(creation_data.mode_definition.name)
        )
        if creation_data.mode_definition
        else '{pipeline_name}.Environment'.format(pipeline_name=pipeline_name),
        fields=remove_none_entries(
            {
                'solids': Field(
                    define_solid_dictionary_cls(
                        '{pipeline_name}.SolidsConfigDictionary'.format(
                            pipeline_name=pipeline_name
                        ),
                        creation_data.solids,
                        creation_data.dependency_structure,
                        creation_data.pipeline_name,
                    )
                ),
                'expectations': Field(
                    define_expectations_config_cls(
                        '{pipeline_name}.ExpectationsConfig'.format(pipeline_name=pipeline_name)
                    )
                ),
                'storage': Field(
                    define_storage_config_cls(
                        '{pipeline_name}.{mode_name}.StorageConfig'.format(
                            pipeline_name=pipeline_name,
                            mode_name=camelcase(creation_data.mode_definition.name),
                        ),
                        creation_data.mode_definition,
                    ),
                    is_optional=True,
                ),
                'execution': Field(
                    define_execution_config_cls(
                        '{pipeline_name}.ExecutionConfig'.format(pipeline_name=pipeline_name)
                    )
                ),
                'loggers': Field(
                    define_logger_dictionary_cls(
                        '{pipeline_name}.LoggerConfig'.format(pipeline_name=pipeline_name),
                        creation_data,
                    )
                ),
                'resources': Field(
                    define_mode_resources_dictionary_cls(
                        pipeline_name, creation_data.mode_definition
                    )
                ),
            }
        ),
    )


def define_expectations_config_cls(name):
    check.str_param(name, 'name')

    return SystemNamedDict(
        name, fields={'evaluate': Field(Bool, is_optional=True, default_value=True)}
    )


def define_storage_config_cls(type_name, mode_definition):
    check.str_param(type_name, 'type_name')
    check.inst_param(mode_definition, 'mode_definition', ModeDefinition)

    fields = {}

    for storage_def in mode_definition.system_storage_defs:
        fields[storage_def.name] = Field(
            SystemNamedDict(
                name='{type_name}.{storage_name}'.format(
                    type_name=type_name, storage_name=camelcase(storage_def.name)
                ),
                fields={'config': storage_def.config_field} if storage_def.config_field else {},
            )
        )

    return SystemNamedSelector(type_name, fields)


def get_inputs_field(solid, handle, dependency_structure, pipeline_name):
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(handle, 'handle', SolidHandle)
    check.inst_param(dependency_structure, 'dependency_structure', DependencyStructure)
    check.str_param(pipeline_name, 'pipeline_name')

    if not solid.definition.has_configurable_inputs:
        return None

    inputs_field_fields = {}
    for name, inp in solid.definition.input_dict.items():
        if inp.runtime_type.input_hydration_config:
            inp_handle = SolidInputHandle(solid, inp)
            # If this input is not satisfied by a dependency you must
            # provide it via config
            if not dependency_structure.has_deps(inp_handle) and not solid.container_maps_input(
                name
            ):
                inputs_field_fields[name] = FieldImpl(
                    inp.runtime_type.input_hydration_config.schema_type
                )

    if not inputs_field_fields:
        return None

    return Field(
        SystemNamedDict(
            '{pipeline_name}.{solid_handle}.Inputs'.format(
                pipeline_name=camelcase(pipeline_name), solid_handle=handle.camelcase()
            ),
            inputs_field_fields,
        )
    )


def get_outputs_field(solid, handle, pipeline_name):
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(handle, 'handle', SolidHandle)
    check.str_param(pipeline_name, 'pipeline_name')

    solid_def = solid.definition

    if not solid_def.has_configurable_outputs:
        return None

    output_dict_fields = {}
    for name, out in solid_def.output_dict.items():
        if out.runtime_type.output_materialization_config:
            output_dict_fields[name] = Field(
                type(out.runtime_type.output_materialization_config.schema_type), is_optional=True
            )

    output_entry_dict = SystemNamedDict(
        '{pipeline_name}.{solid_handle}.Outputs'.format(
            pipeline_name=camelcase(pipeline_name), solid_handle=handle.camelcase()
        ),
        output_dict_fields,
    )

    return Field(List[output_entry_dict], is_optional=True)


def define_isolid_field(solid, handle, dependency_structure, pipeline_name):
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(handle, 'handle', SolidHandle)

    check.str_param(pipeline_name, 'pipeline_name')

    if isinstance(solid.definition, CompositeSolidDefinition):
        composite_def = solid.definition
        child_solids_config_field = Field(
            define_solid_dictionary_cls(
                '{pipeline_name}.CompositeSolidsDict.{solid_handle}'.format(
                    pipeline_name=camelcase(pipeline_name), solid_handle=handle.camelcase()
                ),
                composite_def.solids,
                composite_def.dependency_structure,
                pipeline_name,
                handle,
            )
        )

        composite_config_dict = {
            'inputs': get_inputs_field(solid, handle, dependency_structure, pipeline_name),
            'outputs': get_outputs_field(solid, handle, pipeline_name),
        }

        # Mask solid config for solids beneath this level if config mapping is provided
        if composite_def.has_config_mapping:
            composite_config_dict['config'] = composite_def.config_mapping.config_field
        else:
            composite_config_dict['solids'] = child_solids_config_field

        return Field(
            SolidContainerConfigDict(
                '{name}CompositeSolidConfig'.format(name=str(handle)),
                remove_none_entries(composite_config_dict),
                handle=handle,
                child_solids_config_field=child_solids_config_field,
            )
        )

    elif isinstance(solid.definition, SolidDefinition):
        solid_config_type = define_solid_config_cls(
            '{pipeline_name}.SolidConfig.{solid_handle}'.format(
                pipeline_name=camelcase(pipeline_name), solid_handle=handle.camelcase()
            ),
            solid.definition.config_field,
            inputs_field=get_inputs_field(solid, handle, dependency_structure, pipeline_name),
            outputs_field=get_outputs_field(solid, handle, pipeline_name),
        )
        return Field(solid_config_type)
    else:
        check.invariant(
            'Unexpected ISolidDefinition type {type}'.format(type=type(solid.definition))
        )


def define_solid_dictionary_cls(
    name, solids, dependency_structure, pipeline_name, parent_handle=None
):
    check.str_param(name, 'name')
    check.list_param(solids, 'solids', of_type=Solid)
    check.inst_param(dependency_structure, 'dependency_structure', DependencyStructure)
    check.str_param(pipeline_name, 'pipeline_name')
    check.opt_inst_param(parent_handle, 'parent_handle', SolidHandle)

    fields = {}
    for solid in solids:
        if solid.definition.has_config_entry:
            fields[solid.name] = define_isolid_field(
                solid,
                SolidHandle(solid.name, solid.definition.name, parent_handle),
                dependency_structure,
                pipeline_name,
            )

    return SystemNamedDict(name, fields)


def define_execution_config_cls(name):
    check.str_param(name, 'name')
    return SystemNamedDict(name, {})


def iterate_solid_def_types(solid_def):

    if isinstance(solid_def, SolidDefinition):
        if solid_def.config_field:
            for runtime_type in iterate_config_types(solid_def.config_field.config_type):
                yield runtime_type
    elif isinstance(solid_def, CompositeSolidDefinition):
        for solid in solid_def.solids:
            for def_type in iterate_solid_def_types(solid.definition):
                yield def_type

    else:
        check.invariant('Unexpected ISolidDefinition type {type}'.format(type=type(solid_def)))


def _gather_all_schemas(solid_defs):
    runtime_types = construct_runtime_type_dictionary(solid_defs)
    for rtt in runtime_types.values():
        if rtt.input_hydration_config:
            for ct in iterate_config_types(rtt.input_hydration_config.schema_type):
                yield ct
        if rtt.output_materialization_config:
            for ct in iterate_config_types(rtt.output_materialization_config.schema_type):
                yield ct


def _gather_all_config_types(solid_defs, environment_type):
    check.list_param(solid_defs, 'solid_defs', ISolidDefinition)
    check.inst_param(environment_type, 'environment_type', ConfigType)

    for solid_def in solid_defs:
        for runtime_type in iterate_solid_def_types(solid_def):
            yield runtime_type

    for runtime_type in iterate_config_types(environment_type):
        yield runtime_type


def construct_config_type_dictionary(solid_defs, environment_type):
    check.list_param(solid_defs, 'solid_defs', ISolidDefinition)
    check.inst_param(environment_type, 'environment_type', ConfigType)

    type_dict_by_name = {t.name: t for t in ALL_CONFIG_BUILTINS}
    type_dict_by_key = {t.key: t for t in ALL_CONFIG_BUILTINS}
    all_types = list(_gather_all_config_types(solid_defs, environment_type)) + list(
        _gather_all_schemas(solid_defs)
    )

    for config_type in all_types:
        name = config_type.name
        if name and name in type_dict_by_name:
            if type(config_type) is not type(type_dict_by_name[name]):
                raise DagsterInvalidDefinitionError(
                    (
                        'Type names must be unique. You have constructed two different '
                        'instances of types with the same name "{name}".'
                    ).format(name=name)
                )
        elif name:
            type_dict_by_name[config_type.name] = config_type

        key = config_type.key

        if key in type_dict_by_key:
            if type(config_type) is not type(type_dict_by_key[key]):
                raise DagsterInvalidDefinitionError(
                    (
                        'Type keys must be unique. You have constructed two different '
                        'instances of types with the same key "{key}".'
                    ).format(key=key)
                )

        else:
            type_dict_by_key[config_type.key] = config_type

    return type_dict_by_name, type_dict_by_key



import pytest

from dagster import (
    DagsterInvariantViolationError,
    DagsterStepOutputNotFoundError,
    DependencyDefinition,
    ExpectationResult,
    InputDefinition,
    Output,
    OutputDefinition,
    PipelineDefinition,
    SolidDefinition,
    execute_pipeline,
)
from dagster.core.definitions import IOExpectationDefinition


def test_multiple_outputs():
    def _t_fn(*_args):
        yield Output(output_name='output_one', value='foo')
        yield Output(output_name='output_two', value='bar')

    solid = SolidDefinition(
        name='multiple_outputs',
        input_defs=[],
        output_defs=[OutputDefinition(name='output_one'), OutputDefinition(name='output_two')],
        compute_fn=_t_fn,
    )

    pipeline = PipelineDefinition(solid_defs=[solid])

    result = execute_pipeline(pipeline)
    solid_result = result.solid_result_list[0]

    assert solid_result.solid.name == 'multiple_outputs'
    assert solid_result.output_value('output_one') == 'foo'
    assert solid_result.output_value('output_two') == 'bar'

    with pytest.raises(
        DagsterInvariantViolationError, match='not_defined not defined in solid multiple_outputs'
    ):
        solid_result.output_value('not_defined')


def test_multiple_outputs_expectations():
    called = {}

    def _expect_fn_one(*_args, **_kwargs):
        called['expectation_one'] = True
        return ExpectationResult(success=True)

    def _expect_fn_two(*_args, **_kwargs):
        called['expectation_two'] = True
        return ExpectationResult(success=True)

    def _compute_fn(*_args, **_kwargs):
        yield Output('foo', 'output_one')
        yield Output('bar', 'output_two')

    solid = SolidDefinition(
        name='multiple_outputs',
        input_defs=[],
        output_defs=[
            OutputDefinition(
                name='output_one',
                expectations=[
                    IOExpectationDefinition(name='some_expectation', expectation_fn=_expect_fn_one)
                ],
            ),
            OutputDefinition(
                name='output_two',
                expectations=[
                    IOExpectationDefinition(
                        name='some_other_expectation', expectation_fn=_expect_fn_two
                    )
                ],
            ),
        ],
        compute_fn=_compute_fn,
    )

    pipeline = PipelineDefinition(solid_defs=[solid])

    result = execute_pipeline(pipeline)

    assert result.success
    assert called['expectation_one']
    assert called['expectation_two']


def test_wrong_multiple_output():
    def _t_fn(*_args):
        yield Output(output_name='mismatch', value='foo')

    solid = SolidDefinition(
        name='multiple_outputs',
        input_defs=[],
        output_defs=[OutputDefinition(name='output_one')],
        compute_fn=_t_fn,
    )

    pipeline = PipelineDefinition(solid_defs=[solid])

    with pytest.raises(DagsterInvariantViolationError):
        execute_pipeline(pipeline)


def test_multiple_outputs_of_same_name_disallowed():
    # make this illegal until it is supported

    def _t_fn(*_args):
        yield Output(output_name='output_one', value='foo')
        yield Output(output_name='output_one', value='foo')

    solid = SolidDefinition(
        name='multiple_outputs',
        input_defs=[],
        output_defs=[OutputDefinition(name='output_one')],
        compute_fn=_t_fn,
    )

    pipeline = PipelineDefinition(solid_defs=[solid])

    with pytest.raises(DagsterInvariantViolationError):
        execute_pipeline(pipeline)


def test_multiple_outputs_only_emit_one():
    def _t_fn(*_args):
        yield Output(output_name='output_one', value='foo')

    solid = SolidDefinition(
        name='multiple_outputs',
        input_defs=[],
        output_defs=[
            OutputDefinition(name='output_one'),
            OutputDefinition(name='output_two', is_optional=True),
        ],
        compute_fn=_t_fn,
    )

    called = {}

    def _compute_fn_one(*_args, **_kwargs):
        called['one'] = True

    downstream_one = SolidDefinition(
        name='downstream_one',
        input_defs=[InputDefinition('some_input')],
        output_defs=[],
        compute_fn=_compute_fn_one,
    )

    def _compute_fn_two(*_args, **_kwargs):
        raise Exception('do not call me')

    downstream_two = SolidDefinition(
        name='downstream_two',
        input_defs=[InputDefinition('some_input')],
        output_defs=[],
        compute_fn=_compute_fn_two,
    )

    pipeline = PipelineDefinition(
        solid_defs=[solid, downstream_one, downstream_two],
        dependencies={
            'downstream_one': {'some_input': DependencyDefinition(solid.name, output='output_one')},
            'downstream_two': {'some_input': DependencyDefinition(solid.name, output='output_two')},
        },
    )

    result = execute_pipeline(pipeline)
    assert result.success

    assert called['one']
    solid_result = result.result_for_solid('multiple_outputs')
    assert set(solid_result.output_values.keys()) == set(['output_one'])

    with pytest.raises(
        DagsterInvariantViolationError, match='not_defined not defined in solid multiple_outputs'
    ):
        solid_result.output_value('not_defined')

    with pytest.raises(DagsterInvariantViolationError, match='Did not find result output_two'):
        solid_result.output_value('output_two')

    with pytest.raises(
        DagsterInvariantViolationError,
        match='Try to get result for solid not_present in <<unnamed>>. No such solid.',
    ):
        result.result_for_solid('not_present')

    assert result.result_for_solid('downstream_two').skipped


def test_missing_non_optional_output_fails():
    def _t_fn(*_args):
        yield Output(output_name='output_one', value='foo')

    solid = SolidDefinition(
        name='multiple_outputs',
        input_defs=[],
        output_defs=[OutputDefinition(name='output_one'), OutputDefinition(name='output_two')],
        compute_fn=_t_fn,
    )

    pipeline = PipelineDefinition(solid_defs=[solid])

    with pytest.raises(DagsterStepOutputNotFoundError):
        execute_pipeline(pipeline)



import os
from glob import glob

import six
import yaml

from dagster import check
from dagster.utils.yaml_utils import merge_yamls
from dagster.core.errors import DagsterInvalidDefinitionError, DagsterInvariantViolationError
from .mode import DEFAULT_MODE_NAME


class PresetDefinition:
    '''Defines a preset configuration that a pipeline can execute in. Presets can be used in Dagit to
    load the predefined configuration in to the tool, or in a script or test as follows:

    .. code-block:: python

        execute_pipeline(**pipeline_def.get_preset('example'))


    Args:
        name (str):
            Name of this preset, must be unique amongst presets for a pipeline.
        environment_files (Optional[List[str]]):
            List of paths or glob patterns for yaml files to load and parse as the enivornment
            config for this preset.
        solid_subset (Optional[List[str]]):
            The list of names of solid invocations to execute for this preset.
        mode (Optional[str]):
            The mode to apply when executing this preset. Defaults to 'default'.
    '''

    def __init__(self, name, environment_files=None, solid_subset=None, mode=None):
        self.name = check.str_param(name, 'name')
        self.environment_files = check.opt_list_param(
            environment_files, 'environment_files', of_type=str
        )
        self.solid_subset = check.opt_nullable_list_param(solid_subset, 'solid_subset', of_type=str)
        self.mode = check.opt_str_param(mode, 'mode', DEFAULT_MODE_NAME)

    def get_environment_dict(self, pipeline_name):
        if self.environment_files is None:
            return None

        file_set = set()
        for file_glob in self.environment_files:
            files = glob(file_glob)
            if not files:
                raise DagsterInvalidDefinitionError(
                    'File or glob pattern "{file_glob}" for "environment_files" in preset '
                    '"{name}" for pipline "{pipeline}" produced no results.'.format(
                        name=self.name, file_glob=file_glob, pipeline=pipeline_name
                    )
                )

            file_set.update(map(os.path.realpath, files))

        try:
            merged = merge_yamls(list(file_set))
        except yaml.YAMLError as err:
            six.raise_from(
                DagsterInvariantViolationError(
                    'Encountered error attempting to parse yaml. Parsing files {file_set} '
                    'loaded by file/patterns {files} on preset "{name}" for pipeline "{pipeline}".'.format(
                        file_set=file_set,
                        files=self.environment_files,
                        name=self.name,
                        pipeline=pipeline_name,
                    )
                ),
                err,
            )

        return merged

    def get_environment_yaml(self, pipeline_name):
        merged = self.get_environment_dict(pipeline_name)
        if merged is None:
            return None

        return yaml.dump(merged, default_flow_style=False)



from dagster import check
from dagster.core.definitions import Solid, SolidHandle, SolidInputHandle, SolidOutputHandle, Output

from .objects import (
    ExecutionStep,
    ExecutionValueSubplan,
    SingleOutputStepCreationData,
    StepInput,
    StepKind,
    StepOutput,
    StepOutputHandle,
)

JOIN_OUTPUT = 'join_output'


def __select_first_join(_context, inputs):
    yield Output(output_name=JOIN_OUTPUT, value=list(inputs.values())[0])


def __merge_join(_context, inputs):
    yield Output(output_name=JOIN_OUTPUT, value=list(inputs.values()))


def __empty_join(_context, _inputs):
    yield Output(output_name=JOIN_OUTPUT, value=None)


def create_join_step(pipeline_name, solid, key_suffix, prev_steps, prev_output_name, handle):
    check.str_param(pipeline_name, 'pipeline_name')
    check.inst_param(solid, 'solid', Solid)
    check.str_param(key_suffix, 'key_suffix')
    check.list_param(prev_steps, 'prev_steps', of_type=ExecutionStep)
    check.invariant(len(prev_steps) > 0)
    check.str_param(prev_output_name, 'output_name')

    step_inputs = []
    seen_runtime_type = None
    seen_optionality = None
    for prev_step in prev_steps:
        prev_step_output = prev_step.step_output_named(prev_output_name)

        if seen_runtime_type is None:
            seen_runtime_type = prev_step_output.runtime_type
        else:
            check.invariant(seen_runtime_type == prev_step_output.runtime_type)

        if seen_optionality is None:
            seen_optionality = prev_step_output.optional
        else:
            check.invariant(seen_optionality == prev_step_output.optional)

        output_handle = StepOutputHandle.from_step(prev_step, prev_output_name)

        step_inputs.append(StepInput(prev_step.key, prev_step_output.runtime_type, output_handle))

    return ExecutionStep(
        pipeline_name=pipeline_name,
        key_suffix=key_suffix,
        step_inputs=step_inputs,
        step_outputs=[StepOutput(JOIN_OUTPUT, seen_runtime_type, optional=seen_optionality)],
        compute_fn=__select_first_join,
        kind=StepKind.JOIN,
        solid_handle=handle,
    )


def create_joining_subplan(
    pipeline_name, solid, join_step_key, parallel_steps, parallel_step_output, handle
):
    '''
    This captures a common pattern of fanning out a single value to N steps,
    where each step has similar structure. The strict requirement here is that each step
    must provide an output named the parameters parallel_step_output.

    This takes those steps and then uses a join node to coalesce them so that downstream
    steps can depend on a single output.

    Currently the join step just does a passthrough with no computation. It remains
    to be seen if there should be any work or verification done in this step, especially
    in multi-process environments that require persistence between steps.
    '''
    check.str_param(pipeline_name, 'pipeline_name')
    check.inst_param(solid, 'solid', Solid)
    check.str_param(join_step_key, 'join_step_key')
    check.list_param(parallel_steps, 'parallel_steps', of_type=ExecutionStep)
    check.str_param(parallel_step_output, 'parallel_step_output')
    check.opt_inst_param(handle, 'handle', SolidHandle)

    for parallel_step in parallel_steps:
        check.invariant(parallel_step.has_step_output(parallel_step_output))

    join_step = create_join_step(
        pipeline_name, solid, join_step_key, parallel_steps, parallel_step_output, handle
    )

    output_name = join_step.step_outputs[0].name
    return ExecutionValueSubplan(
        parallel_steps + [join_step], StepOutputHandle.from_step(join_step, output_name)
    )


def create_join_outputs_step(
    pipeline_name, input_handle, solid_output_handles, step_output_handles, handle
):
    check.str_param(pipeline_name, 'pipeline_name')
    check.inst_param(input_handle, 'input_handle', SolidInputHandle)
    check.list_param(solid_output_handles, 'output_handles', of_type=SolidOutputHandle)
    check.dict_param(
        step_output_handles,
        'step_output_handles',
        key_type=SolidOutputHandle,
        value_type=StepOutputHandle,
    )
    check.opt_inst_param(handle, 'handle', SolidHandle)

    step_inputs = [
        StepInput(
            "{solid_output_handle.solid.name}.{solid_output_handle.output_def.name}".format(
                solid_output_handle=solid_output_handle
            ),
            solid_output_handle.output_def.runtime_type,
            step_output_handles[solid_output_handle],
        )
        for solid_output_handle in solid_output_handles
    ]

    input_type = input_handle.input_def.runtime_type

    join = ExecutionStep(
        pipeline_name=pipeline_name,
        key_suffix='input.{name}.join'.format(name=input_handle.input_def.name),
        step_inputs=step_inputs,
        step_outputs=[StepOutput(JOIN_OUTPUT, input_type, optional=False)],
        compute_fn=__empty_join if input_type.is_nothing else __merge_join,
        kind=StepKind.JOIN,
        solid_handle=handle,
    )

    return SingleOutputStepCreationData(join, JOIN_OUTPUT)



from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError
from .config import ConfigType, ConfigTypeAttributes, DEFAULT_TYPE_ATTRIBUTES
from .default_applier import apply_default_values


def all_optional_type(config_type):
    check.inst_param(config_type, 'config_type', ConfigType)

    if config_type.is_composite or config_type.is_selector:
        for field in config_type.fields.values():
            if not field.is_optional:
                return False
        return True

    return False


class __FieldValueSentinel:
    pass


class __InferOptionalCompositeFieldSentinel:
    pass


FIELD_NO_DEFAULT_PROVIDED = __FieldValueSentinel

INFER_OPTIONAL_COMPOSITE_FIELD = __InferOptionalCompositeFieldSentinel


def check_using_facing_field_param(obj, param_name, error_context_str):
    check.str_param(param_name, 'param_name')
    check.str_param(error_context_str, 'error_context_str')

    if isinstance(obj, FieldImpl):
        return obj

    from dagster.core.types.field import resolve_to_config_type
    from .type_printer import print_config_type_to_string

    config_type = resolve_to_config_type(obj)
    if config_type:
        raise DagsterInvalidDefinitionError(
            (
                'You have passed a config type "{printed_type}" in the parameter '
                '"{param_name}" {error_context_str}. '
                'You have likely forgot to wrap this type in a Field.'
            ).format(
                printed_type=print_config_type_to_string(config_type, with_lines=False),
                error_context_str=error_context_str,
                param_name=param_name,
            )
        )
    else:
        raise DagsterInvalidDefinitionError(
            (
                'You have passed an object {value_repr} of incorrect type '
                '"{type_name}" in the parameter "{param_name}" '
                '{error_context_str} where a Field was expected.'
            ).format(
                error_context_str=error_context_str,
                param_name=param_name,
                value_repr=repr(obj),
                type_name=type(obj).__name__,
            )
        )


def check_user_facing_opt_field_param(obj, param_name, error_context_str):
    check.str_param(param_name, 'param_name')
    check.str_param(error_context_str, 'error_context_str')
    if obj is None:
        return None
    return check_using_facing_field_param(obj, param_name, error_context_str)


def check_field_param(obj, param_name):
    return check.inst_param(obj, param_name, FieldImpl)


def check_opt_field_param(obj, param_name):
    return check.opt_inst_param(obj, param_name, FieldImpl)


class FieldImpl:
    '''
    A field in a config object.

    Attributes:
        config_type (ConfigType): The type of the field.
        default_value (Any):
            If the Field is optional, a default value can be provided when the field value
            is not specified.
        is_optional (bool): Is the field optional.
        description (str): Description of the field.
    '''

    def __init__(
        self,
        config_type,
        default_value=FIELD_NO_DEFAULT_PROVIDED,
        is_optional=INFER_OPTIONAL_COMPOSITE_FIELD,
        is_secret=False,
        description=None,
    ):
        self.config_type = check.inst_param(config_type, 'config_type', ConfigType)

        self.description = check.opt_str_param(description, 'description')
        if is_optional == INFER_OPTIONAL_COMPOSITE_FIELD:
            is_optional = all_optional_type(self.config_type)
            if is_optional is True:
                self._default_value = apply_default_values(self.config_type, None)
            else:
                self._default_value = default_value
        else:
            is_optional = check.bool_param(is_optional, 'is_optional')
            self._default_value = default_value

        if is_optional is False:
            check.param_invariant(
                default_value == FIELD_NO_DEFAULT_PROVIDED,
                'default_value',
                'required arguments should not specify default values',
            )

        if config_type.is_scalar and self._default_value != FIELD_NO_DEFAULT_PROVIDED:
            check.param_invariant(
                config_type.is_config_scalar_valid(self._default_value),
                'default_value',
                'default value not valid for config type {name}, got value {val} of type {type}'.format(
                    name=config_type.name, val=self._default_value, type=type(self._default_value)
                ),
            )

        self.is_optional = is_optional

        self.is_secret = check.bool_param(is_secret, 'is_secret')

    @property
    def is_required(self):
        return not self.is_optional

    @property
    def default_provided(self):
        '''Was a default value provided

        Returns:
            bool: Yes or no
        '''
        return self._default_value != FIELD_NO_DEFAULT_PROVIDED

    @property
    def default_value(self):
        check.invariant(self.default_provided, 'Asking for default value when none was provided')

        if callable(self._default_value):
            return self._default_value()

        return self._default_value

    @property
    def default_value_as_str(self):
        check.invariant(self.default_provided, 'Asking for default value when none was provided')

        if callable(self._default_value):
            return repr(self._default_value)

        return str(self._default_value)

    def __repr__(self):
        return (
            'Field({config_type}, default={default}, is_optional={is_optional}, '
            'is_secret={is_secret})'
        ).format(
            config_type=self.config_type,
            default='@'
            if self._default_value == FIELD_NO_DEFAULT_PROVIDED
            else self._default_value,
            is_optional=self.is_optional,
            is_secret=self.is_secret,
        )


class _ConfigHasFields(ConfigType):
    def __init__(self, fields, *args, **kwargs):
        self.fields = check.dict_param(fields, 'fields', key_type=str, value_type=FieldImpl)
        super(_ConfigHasFields, self).__init__(*args, **kwargs)

    @property
    def inner_types(self):
        return list(set(self._yield_inner_types()))

    def _yield_inner_types(self):
        for field in self.fields.values():
            yield field.config_type
            for inner_type in field.config_type.inner_types:
                yield inner_type


class _ConfigComposite(_ConfigHasFields):
    @property
    def is_composite(self):
        return True

    @property
    def is_permissive_composite(self):
        return False


class _CompositeSolidConfigComposite(_ConfigComposite):
    @property
    def handle(self):
        return self.type_attributes.handle


class _ConfigSelector(_ConfigHasFields):
    @property
    def is_selector(self):
        return True


# HACK HACK HACK
#
# This is not good and a better solution needs to be found. In order
# for the client-side typeahead in dagit to work as currently structured,
# dictionaries need names. While we deal with that we're going to automatically
# name dictionaries. This will cause odd behavior and bugs is you restart
# the server-side process, the type names changes, and you do not refresh the client.
#
# A possible short term mitigation would to name the dictionary based on the hash
# of its member fields to provide stability in between process restarts.
#
class DictCounter:
    _count = 0

    @staticmethod
    def get_next_count():
        DictCounter._count += 1
        return DictCounter._count


def check_user_facing_fields_dict(fields, type_name_msg):
    check.dict_param(fields, 'fields', key_type=str)
    check.str_param(type_name_msg, 'type_name_msg')

    sorted_field_names = sorted(list(fields.keys()))
    for field_name, potential_field in fields.items():
        check_using_facing_field_param(
            potential_field,
            'fields',
            (
                'and it is in the "{field_name}" entry of that dict. It is from '
                'a {type_name_msg} with fields {field_names}'
            ).format(
                type_name_msg=type_name_msg, field_name=field_name, field_names=sorted_field_names
            ),
        )


def NamedDict(name, fields, description=None, type_attributes=DEFAULT_TYPE_ATTRIBUTES):
    '''
    A :py:class:`Dict` with a name allowing it to be referenced by that name.
    '''
    check_user_facing_fields_dict(fields, 'NamedDict named "{}"'.format(name))

    class _NamedDict(_ConfigComposite):
        def __init__(self):
            super(_NamedDict, self).__init__(
                key=name,
                name=name,
                fields=fields,
                description=description,
                type_attributes=type_attributes,
            )

    return _NamedDict


def Dict(fields):
    '''
    Schema for configuration data with string keys and typed values via :py:class:`Field` .

    Args:
        fields (Dict[str, Field])
    '''
    check_user_facing_fields_dict(fields, 'Dict')

    class _Dict(_ConfigComposite):
        def __init__(self):
            key = 'Dict.' + str(DictCounter.get_next_count())
            super(_Dict, self).__init__(
                name=None,
                key=key,
                fields=fields,
                description='A configuration dictionary with typed fields',
                type_attributes=ConfigTypeAttributes(is_builtin=True),
            )

    return _Dict


def PermissiveDict(fields=None):
    '''A permissive dict will permit the user to partially specify the permitted fields. Any fields
    that are specified and passed in will be type checked. Other fields will be allowed, but
    will be ignored by the type checker.
    '''

    if fields:
        check_user_facing_fields_dict(fields, 'PermissiveDict')

    class _PermissiveDict(_ConfigComposite):
        def __init__(self):
            key = 'PermissiveDict.' + str(DictCounter.get_next_count())
            super(_PermissiveDict, self).__init__(
                name=None,
                key=key,
                fields=fields or dict(),
                description='A configuration dictionary with typed fields',
                type_attributes=ConfigTypeAttributes(is_builtin=True),
            )

        @property
        def is_permissive_composite(self):
            return True

    return _PermissiveDict


def Selector(fields):
    '''Selectors are used when you want to be able present several different options to the user but
    force them to select one. For example, it would not make much sense to allow them
    to say that a single input should be sourced from a csv and a parquet file: They must choose.

    Note that in other type systems this might be called an "input union."

    Args:
        fields (Dict[str, Field]):
    '''

    check_user_facing_fields_dict(fields, 'Selector')

    class _Selector(_ConfigSelector):
        def __init__(self):
            key = 'Selector.' + str(DictCounter.get_next_count())
            super(_Selector, self).__init__(
                key=key,
                name=None,
                fields=fields,
                # description='A configuration dictionary with typed fields',
                type_attributes=ConfigTypeAttributes(is_builtin=True),
            )

    return _Selector


def NamedSelector(name, fields, description=None, type_attributes=DEFAULT_TYPE_ATTRIBUTES):
    '''
    A :py:class`Selector` with a name, allowing it to be referenced by that name.

    Args:
        name (str):
        fields (Dict[str, Field])
    '''
    check.str_param(name, 'name')
    check_user_facing_fields_dict(fields, 'NamedSelector named "{}"'.format(name))

    class _NamedSelector(_ConfigSelector):
        def __init__(self):
            super(_NamedSelector, self).__init__(
                key=name,
                name=name,
                fields=fields,
                description=description,
                type_attributes=type_attributes,
            )

    return _NamedSelector



import pytest

from dagster import (
    Dict,
    Field,
    DagsterInvalidConfigError,
    PipelineDefinition,
    SolidDefinition,
    String,
    execute_pipeline,
)


def test_basic_solid_with_config():
    did_get = {}

    def _t_fn(context, _inputs):
        did_get['yep'] = context.solid_config

    solid = SolidDefinition(
        name='solid_with_context',
        input_defs=[],
        output_defs=[],
        config_field=Field(Dict({'some_config': Field(String)})),
        compute_fn=_t_fn,
    )

    pipeline = PipelineDefinition(solid_defs=[solid])

    execute_pipeline(
        pipeline, {'solids': {'solid_with_context': {'config': {'some_config': 'foo'}}}}
    )

    assert 'yep' in did_get
    assert 'some_config' in did_get['yep']


def test_config_arg_mismatch():
    def _t_fn(*_args):
        raise Exception('should not reach')

    solid = SolidDefinition(
        name='solid_with_context',
        input_defs=[],
        output_defs=[],
        config_field=Field(Dict({'some_config': Field(String)})),
        compute_fn=_t_fn,
    )

    pipeline = PipelineDefinition(solid_defs=[solid])

    with pytest.raises(DagsterInvalidConfigError):
        execute_pipeline(
            pipeline, {'solids': {'solid_with_context': {'config': {'some_config': 1}}}}
        )


def test_solid_not_found():
    def _t_fn(*_args):
        raise Exception('should not reach')

    solid = SolidDefinition(name='find_me_solid', input_defs=[], output_defs=[], compute_fn=_t_fn)

    pipeline = PipelineDefinition(solid_defs=[solid])

    with pytest.raises(DagsterInvalidConfigError):
        execute_pipeline(pipeline, {'solids': {'not_found': {'config': {'some_config': 1}}}})


def test_config_for_no_config():
    def _t_fn(*_args):
        raise Exception('should not reach')

    solid_def = SolidDefinition(
        name='no_config_solid', input_defs=[], output_defs=[], compute_fn=_t_fn
    )

    pipeline = PipelineDefinition(solid_defs=[solid_def])

    with pytest.raises(DagsterInvalidConfigError):
        execute_pipeline(pipeline, {'solids': {'no_config_solid': {'config': {'some_config': 1}}}})



from collections import namedtuple
from dagster import check
from dagster.core.types.runtime import RuntimeType, resolve_to_runtime_type

from .expectation import IOExpectationDefinition
from .utils import check_valid_name, DEFAULT_OUTPUT


class OutputDefinition(object):
    '''An OutputDefinition represents an output from a solid. Solids can have multiple
    outputs. In those cases the outputs must be named. Frequently solids have only one
    output, and so the user can construct a single OutputDefinition that will have
    the default name of "result".

    Args:
        dagster_type (DagsterType):
            Type of the output. Defaults to :py:class:`Any` . Basic python types will be
            mapped to the appropriate DagsterType.
        name (Optional[str]): Name of the output. Defaults to "result".
        expectations List[IOExpectationDefinition]:
            **Deprecated**: Expectations for this output.

            Prefer yielding :py:class:`ExpectationResult` directly from solid compute function.
        description (str): Description of the output. Optional.
        is_optional (bool): If this output is optional. Optional, defaults to false.
    '''

    def __init__(
        self, dagster_type=None, name=None, expectations=None, description=None, is_optional=False
    ):
        self.name = check_valid_name(check.opt_str_param(name, 'name', DEFAULT_OUTPUT))

        self.runtime_type = check.inst(resolve_to_runtime_type(dagster_type), RuntimeType)

        self.expectations = check.opt_list_param(
            expectations, 'expectations', of_type=IOExpectationDefinition
        )
        self.description = check.opt_str_param(description, 'description')

        self.optional = check.bool_param(is_optional, 'is_optional')

    @property
    def descriptive_key(self):
        return 'output'

    def mapping_from(self, solid_name, output_name=None):
        return OutputMapping(self, solid_name, output_name)


class OutputMapping(namedtuple('_OutputMapping', 'definition solid_name output_name')):
    def __new__(cls, definition, solid_name, output_name=None):
        return super(OutputMapping, cls).__new__(
            cls,
            check.inst_param(definition, 'definition', OutputDefinition),
            check.str_param(solid_name, 'solid_name'),
            check.opt_str_param(output_name, 'output_name', DEFAULT_OUTPUT),
        )



from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError

from .builtin_enum import BuiltinEnum
from .wrapping import WrappingType


MAGIC_RUNTIME_TYPE_NAME = '__runtime_type'


def is_runtime_type_decorated_klass(klass):
    check.type_param(klass, 'klass')
    return hasattr(klass, MAGIC_RUNTIME_TYPE_NAME)


def check_dagster_type_param(dagster_type, param_name, base_type):

    # Cannot check base_type because of circular references and no fwd declarations
    if dagster_type is None:
        return dagster_type
    if BuiltinEnum.contains(dagster_type):
        return dagster_type
    if isinstance(dagster_type, WrappingType):
        return dagster_type

    if not isinstance(dagster_type, type):
        raise DagsterInvalidDefinitionError(
            'Invalid type for "{param_name}": dagster_type is not an instance of "type", got {dagster_type}'.format(
                param_name=param_name, dagster_type=dagster_type
            )
        )

    if is_runtime_type_decorated_klass(dagster_type):
        return dagster_type

    if not issubclass(dagster_type, base_type):
        raise DagsterInvalidDefinitionError(
            (
                'Parameter {param_name} must be a valid dagster type: A builtin (e.g. String, Int, '
                'etc), a wrapping type (List or Optional), or a type class. Got {dagster_type}'
            ).format(param_name=param_name, dagster_type=dagster_type)
        )

    return dagster_type



from functools import partial
import six

from dagster import check

from dagster.core.storage.type_storage import TypeStoragePlugin

from .builtin_enum import BuiltinEnum
from .builtin_config_schemas import BuiltinSchemas

from .config import ConfigType
from .config import List as ConfigList
from .config import Nullable as ConfigNullable

from .config_schema import InputHydrationConfig, OutputMaterializationConfig

from .marshal import SerializationStrategy, PickleSerializationStrategy
from .dagster_type import check_dagster_type_param
from .wrapping import WrappingListType, WrappingNullableType


def check_opt_config_cls_param(config_cls, param_name):
    if config_cls is None:
        return config_cls
    check.invariant(isinstance(config_cls, type))
    check.param_invariant(issubclass(config_cls, ConfigType), param_name)
    return config_cls


class RuntimeType(object):
    '''
    The class backing DagsterTypes as they are used during execution.
    '''

    def __init__(
        self,
        key,
        name,
        is_builtin=False,
        description=None,
        input_hydration_config=None,
        output_materialization_config=None,
        serialization_strategy=None,
        auto_plugins=None,
    ):

        type_obj = type(self)
        if type_obj in RuntimeType.__cache:
            check.failed(
                (
                    '{type_obj} already in cache. You **must** use the inst() class method '
                    'to construct RuntimeType and not the ctor'.format(type_obj=type_obj)
                )
            )

        self.key = check.str_param(key, 'key')
        self.name = check.opt_str_param(name, 'name')
        self.description = check.opt_str_param(description, 'description')
        self.input_hydration_config = check.opt_inst_param(
            input_hydration_config, 'input_hydration_config', InputHydrationConfig
        )
        self.output_materialization_config = check.opt_inst_param(
            output_materialization_config,
            'output_materialization_config',
            OutputMaterializationConfig,
        )
        self.serialization_strategy = check.opt_inst_param(
            serialization_strategy,
            'serialization_strategy',
            SerializationStrategy,
            PickleSerializationStrategy(),
        )

        auto_plugins = check.opt_list_param(auto_plugins, 'auto_plugins', of_type=type)

        check.param_invariant(
            all(
                issubclass(auto_plugin_type, TypeStoragePlugin) for auto_plugin_type in auto_plugins
            ),
            'auto_plugins',
        )

        self.auto_plugins = auto_plugins

        self.is_builtin = check.bool_param(is_builtin, 'is_builtin')

    __cache = {}

    @classmethod
    def inst(cls):
        if cls not in RuntimeType.__cache:
            RuntimeType.__cache[cls] = cls()  # pylint: disable=E1120
        return RuntimeType.__cache[cls]

    @staticmethod
    def from_builtin_enum(builtin_enum):
        check.invariant(BuiltinEnum.contains(builtin_enum), 'must be member of BuiltinEnum')
        return _RUNTIME_MAP[builtin_enum]

    @property
    def display_name(self):
        return self.name

    def type_check(self, value):
        pass

    @property
    def is_any(self):
        return False

    @property
    def is_scalar(self):
        return False

    @property
    def is_list(self):
        return False

    @property
    def is_nullable(self):
        return False

    @property
    def inner_types(self):
        return []

    @property
    def is_nothing(self):
        return False


class BuiltinScalarRuntimeType(RuntimeType):
    def __init__(self, *args, **kwargs):
        name = type(self).__name__
        super(BuiltinScalarRuntimeType, self).__init__(
            key=name, name=name, is_builtin=True, *args, **kwargs
        )

    @property
    def is_scalar(self):
        return True


class Int(BuiltinScalarRuntimeType):
    def __init__(self):
        super(Int, self).__init__(
            input_hydration_config=BuiltinSchemas.INT_INPUT,
            output_materialization_config=BuiltinSchemas.INT_OUTPUT,
        )

    def type_check(self, value):
        from dagster.core.definitions.events import Failure

        if not isinstance(value, six.integer_types):
            raise Failure(_typemismatch_error_str(value, 'int'))


def _typemismatch_error_str(value, expected_type_desc):
    return 'Value "{value}" of python type "{python_type}" must be a {type_desc}.'.format(
        value=value, python_type=type(value).__name__, type_desc=expected_type_desc
    )


def _throw_if_not_string(value):
    from dagster.core.definitions.events import Failure

    if not isinstance(value, six.string_types):
        raise Failure(_typemismatch_error_str(value, 'string'))


class String(BuiltinScalarRuntimeType):
    def __init__(self):
        super(String, self).__init__(
            input_hydration_config=BuiltinSchemas.STRING_INPUT,
            output_materialization_config=BuiltinSchemas.STRING_OUTPUT,
        )

    def type_check(self, value):
        _throw_if_not_string(value)


class Path(BuiltinScalarRuntimeType):
    def __init__(self):
        super(Path, self).__init__(
            input_hydration_config=BuiltinSchemas.PATH_INPUT,
            output_materialization_config=BuiltinSchemas.PATH_OUTPUT,
        )

    def type_check(self, value):
        _throw_if_not_string(value)


class Float(BuiltinScalarRuntimeType):
    def __init__(self):
        super(Float, self).__init__(
            input_hydration_config=BuiltinSchemas.FLOAT_INPUT,
            output_materialization_config=BuiltinSchemas.FLOAT_OUTPUT,
        )

    def type_check(self, value):
        from dagster.core.definitions.events import Failure

        if not isinstance(value, float):
            raise Failure(_typemismatch_error_str(value, 'float'))


class Bool(BuiltinScalarRuntimeType):
    def __init__(self):
        super(Bool, self).__init__(
            input_hydration_config=BuiltinSchemas.BOOL_INPUT,
            output_materialization_config=BuiltinSchemas.BOOL_OUTPUT,
        )

    def type_check(self, value):
        from dagster.core.definitions.events import Failure

        if not isinstance(value, bool):
            raise Failure(_typemismatch_error_str(value, 'bool'))


class Anyish(RuntimeType):
    def __init__(
        self,
        key,
        name,
        input_hydration_config=None,
        output_materialization_config=None,
        is_builtin=False,
        description=None,
    ):
        super(Anyish, self).__init__(
            key=key,
            name=name,
            input_hydration_config=input_hydration_config,
            output_materialization_config=output_materialization_config,
            is_builtin=is_builtin,
            description=description,
        )

    @property
    def is_any(self):
        return True


class Any(Anyish):
    def __init__(self):
        super(Any, self).__init__(
            key='Any',
            name='Any',
            input_hydration_config=BuiltinSchemas.ANY_INPUT,
            output_materialization_config=BuiltinSchemas.ANY_OUTPUT,
            is_builtin=True,
        )


def define_any_type(name, description=None):
    class NamedAnyType(Anyish):
        def __init__(self):
            super(NamedAnyType, self).__init__(key=name, name=name, description=description)

    return NamedAnyType


class Nothing(RuntimeType):
    def __init__(self):
        super(Nothing, self).__init__(
            key='Nothing',
            name='Nothing',
            input_hydration_config=None,
            output_materialization_config=None,
            is_builtin=True,
        )

    @property
    def is_nothing(self):
        return True

    def type_check(self, value):
        from dagster.core.definitions.events import Failure

        if value is not None:
            raise Failure('Value {value} must be None.')


class PythonObjectType(RuntimeType):
    def __init__(self, python_type, key=None, name=None, typecheck_metadata_fn=None, **kwargs):
        name = check.opt_str_param(name, 'name', type(self).__name__)
        key = check.opt_str_param(key, 'key', name)
        super(PythonObjectType, self).__init__(key=key, name=name, **kwargs)
        self.python_type = check.type_param(python_type, 'python_type')
        self.typecheck_metadata_fn = check.opt_callable_param(
            typecheck_metadata_fn, 'typecheck_metadata_fn'
        )

    def type_check(self, value):
        from dagster.core.definitions.events import Failure

        if not isinstance(value, self.python_type):
            raise Failure(
                'Value {value} should be of type {type_name}.'.format(
                    value=value, type_name=self.python_type.__name__
                )
            )

        if self.typecheck_metadata_fn:
            return self.typecheck_metadata_fn(value)


def define_python_dagster_type(
    python_type,
    name=None,
    description=None,
    input_hydration_config=None,
    output_materialization_config=None,
    serialization_strategy=None,
    auto_plugins=None,
    typecheck_metadata_fn=None,
):
    '''
    The dagster typesystem is very flexible, and the body of a typecheck can be
    a function that does *anything*. (For that level of flexiblity one should inherit
    from RuntimeType directly)  However its very common to want to generate a dagster
    type whose only typecheck is against a python type:

    DateTime = define_python_dagster_type(datetime.datetime, name='DateTime')

    Args:
        python_type (cls)
            The python type you want check against.
        name (Optional[str]): 
            Name of the dagster type. Defaults to the name of the python_type. 
        description (Optiona[str]):
        input_hydration_config (Optional[InputHydrationConfig]):
            An instance of a class that inherits from :py:class:`InputHydrationConfig` that
            can map config data to a value of this type.

        output_materialization_config (Optiona[OutputMaterializationConfig]):
            An instance of a class that inherits from :py:class:`OutputMaterializationConfig` that
            can map config data to persisting values of this type.

        serialization_strategy (Optional[SerializationStrategy]):
            The default behavior for how to serialize this value for persisting between execution
            steps.

        auto_plugins (Optional[List[type]]):
            types *must* subclass from TypeStoragePlugin.
            This allows for types to specify serialization that depends on what storage
            is being used to serialize intermediates. In these cases the serialization_strategy
            is not sufficient because serialization requires specialized API calls, e.g.
            to call an s3 API directly instead of using a generic file object. See
            dagster_pyspark.DataFrame for an example of auto_plugins.

        typecheck_metadata_fn (Callable):
            It is used to emit metadata when you successfully check a type. This allows
            the user specifiy that function that emits that metadata object whenever the typecheck
            succeeds. The passed in function takes the value being evaluated and returns a
            TypeCheck event.

            See dagster_pandas.DataFrame for an example

    '''
    check.type_param(python_type, 'python_type')
    check.opt_str_param(name, 'name')
    check.opt_str_param(description, 'description')
    check.opt_inst_param(input_hydration_config, 'input_hydration_config', InputHydrationConfig)
    check.opt_inst_param(
        output_materialization_config, 'output_materialization_config', OutputMaterializationConfig
    )
    check.opt_inst_param(
        serialization_strategy,
        'serialization_strategy',
        SerializationStrategy,
        default=PickleSerializationStrategy(),
    )

    auto_plugins = check.opt_list_param(auto_plugins, 'auto_plugins', of_type=type)
    check.param_invariant(
        all(issubclass(auto_plugin_type, TypeStoragePlugin) for auto_plugin_type in auto_plugins),
        'auto_plugins',
    )

    check.opt_callable_param(typecheck_metadata_fn, 'typecheck_metadata_fn')

    class _ObjectType(PythonObjectType):
        def __init__(self):
            super(_ObjectType, self).__init__(
                python_type=python_type,
                name=name,
                description=description,
                input_hydration_config=input_hydration_config,
                output_materialization_config=output_materialization_config,
                serialization_strategy=serialization_strategy,
                auto_plugins=auto_plugins,
                typecheck_metadata_fn=typecheck_metadata_fn,
            )

    return _ObjectType


def _create_nullable_input_schema(inner_type):
    if not inner_type.input_hydration_config:
        return None

    nullable_type = ConfigNullable(inner_type.input_hydration_config.schema_type).inst()

    class _NullableSchema(InputHydrationConfig):
        @property
        def schema_type(self):
            return nullable_type

        def construct_from_config_value(self, context, config_value):
            if config_value is None:
                return None
            return inner_type.input_hydration_config.construct_from_config_value(
                context, config_value
            )

    return _NullableSchema()


class NullableType(RuntimeType):
    def __init__(self, inner_type):
        key = 'Optional.' + inner_type.key
        super(NullableType, self).__init__(
            key=key, name=None, input_hydration_config=_create_nullable_input_schema(inner_type)
        )
        self.inner_type = inner_type

    @property
    def display_name(self):
        return self.inner_type.display_name + '?'

    def type_check(self, value):
        return None if value is None else self.inner_type.type_check(value)

    @property
    def is_nullable(self):
        return True

    @property
    def inner_types(self):
        return [self.inner_type] + self.inner_type.inner_types


def _create_list_input_schema(inner_type):
    if not inner_type.input_hydration_config:
        return None

    list_type = ConfigList(inner_type.input_hydration_config.schema_type).inst()

    class _ListSchema(InputHydrationConfig):
        @property
        def schema_type(self):
            return list_type

        def construct_from_config_value(self, context, config_value):
            convert_item = partial(
                inner_type.input_hydration_config.construct_from_config_value, context
            )
            return list(map(convert_item, config_value))

    return _ListSchema()


class ListType(RuntimeType):
    def __init__(self, inner_type):
        key = 'List.' + inner_type.key
        super(ListType, self).__init__(
            key=key, name=None, input_hydration_config=_create_list_input_schema(inner_type)
        )
        self.inner_type = inner_type

    @property
    def display_name(self):
        return '[' + self.inner_type.display_name + ']'

    def type_check(self, value):
        from dagster.core.definitions.events import Failure

        if not isinstance(value, list):
            raise Failure('Value must be a list, got {value}'.format(value=value))

        for item in value:
            self.inner_type.type_check(item)

    @property
    def is_list(self):
        return True

    @property
    def inner_types(self):
        return [self.inner_type] + self.inner_type.inner_types


def Optional(inner_type):
    check.inst_param(inner_type, 'inner_type', RuntimeType)

    class _Nullable(NullableType):
        def __init__(self):
            super(_Nullable, self).__init__(inner_type)

    return _Nullable.inst()


def List(inner_type):
    check.inst_param(inner_type, 'inner_type', RuntimeType)

    class _List(ListType):
        def __init__(self):
            super(_List, self).__init__(inner_type)

    return _List.inst()


class Stringish(RuntimeType):
    def __init__(self, key=None, name=None, **kwargs):
        name = check.opt_str_param(name, 'name', type(self).__name__)
        key = check.opt_str_param(key, 'key', name)
        super(Stringish, self).__init__(key=key, name=name, **kwargs)

    def is_scalar(self):
        return True

    def type_check(self, value):
        return _throw_if_not_string(value)


_RUNTIME_MAP = {
    BuiltinEnum.ANY: Any.inst(),
    BuiltinEnum.BOOL: Bool.inst(),
    BuiltinEnum.FLOAT: Float.inst(),
    BuiltinEnum.INT: Int.inst(),
    BuiltinEnum.PATH: Path.inst(),
    BuiltinEnum.STRING: String.inst(),
    BuiltinEnum.NOTHING: Nothing.inst(),
}


def resolve_to_runtime_type(dagster_type):
    # circular dep
    from .decorator import is_runtime_type_decorated_klass, get_runtime_type_on_decorated_klass
    from .mapping import remap_python_type

    dagster_type = remap_python_type(dagster_type)

    check_dagster_type_param(dagster_type, 'dagster_type', RuntimeType)

    if dagster_type is None:
        return Any.inst()
    if BuiltinEnum.contains(dagster_type):
        return RuntimeType.from_builtin_enum(dagster_type)
    if isinstance(dagster_type, WrappingListType):
        return resolve_to_runtime_list(dagster_type)
    if isinstance(dagster_type, WrappingNullableType):
        return resolve_to_runtime_nullable(dagster_type)
    if is_runtime_type_decorated_klass(dagster_type):
        return get_runtime_type_on_decorated_klass(dagster_type)
    if issubclass(dagster_type, RuntimeType):
        return dagster_type.inst()

    check.failed('should not reach')


def resolve_to_runtime_list(list_type):
    check.inst_param(list_type, 'list_type', WrappingListType)
    return List(resolve_to_runtime_type(list_type.inner_type))


def resolve_to_runtime_nullable(nullable_type):
    check.inst_param(nullable_type, 'nullable_type', WrappingNullableType)
    return Optional(resolve_to_runtime_type(nullable_type.inner_type))


ALL_RUNTIME_BUILTINS = set(_RUNTIME_MAP.values())


def construct_runtime_type_dictionary(solid_defs):
    type_dict = {t.name: t for t in ALL_RUNTIME_BUILTINS}
    for solid_def in solid_defs:
        for runtime_type in solid_def.all_runtime_types():
            type_dict[runtime_type.name] = runtime_type

    return type_dict



from dagster.core.definitions import (
    CompositeSolidDefinition,
    DependencyDefinition,
    EventMetadataEntry,
    TextMetadataEntryData,
    UrlMetadataEntryData,
    PathMetadataEntryData,
    JsonMetadataEntryData,
    ExecutionTargetHandle,
    ExpectationResult,
    Failure,
    InputDefinition,
    LoggerDefinition,
    Materialization,
    ModeDefinition,
    MultiDependencyDefinition,
    OutputDefinition,
    PipelineDefinition,
    PresetDefinition,
    RepositoryDefinition,
    ResourceDefinition,
    Output,
    SolidDefinition,
    SolidInvocation,
    SystemStorageDefinition,
    SystemStorageData,
    TypeCheck,
    composite_solid,
    lambda_solid,
    logger,
    pipeline,
    resource,
    solid,
    system_storage,
)

from dagster.core.events import DagsterEventType

from dagster.core.execution.api import (
    execute_pipeline,
    execute_pipeline_iterator,
    execute_pipeline_with_preset,
)

from dagster.core.execution.config import (
    InProcessExecutorConfig,
    MultiprocessExecutorConfig,
    RunConfig,
)

from dagster.core.execution.context.init import InitResourceContext

from dagster.core.execution.context.logger import InitLoggerContext

from dagster.core.execution.plan.objects import StepKind

from dagster.core.execution.results import PipelineExecutionResult, SolidExecutionResult

from dagster.core.errors import (
    DagsterExecutionStepExecutionError,
    DagsterExecutionStepNotFoundError,
    DagsterInvalidConfigError,
    DagsterInvalidDefinitionError,
    DagsterInvariantViolationError,
    DagsterResourceFunctionError,
    DagsterStepOutputNotFoundError,
    DagsterTypeCheckError,
    DagsterUserCodeExecutionError,
)

from dagster.core.storage.init import InitSystemStorageContext
from dagster.core.storage.file_manager import FileHandle, LocalFileHandle

from dagster.core.system_config.objects import EnvironmentConfig

from dagster.core.types import (
    Any,
    Bool,
    Dict,
    Field,
    Float,
    Int,
    List,
    NamedDict,
    Nothing,
    Optional,
    Path,
    PermissiveDict,
    String,
    input_hydration_config,
    output_materialization_config,
)

from dagster.core.types.config import ConfigType, ConfigScalar, Enum, EnumValue

from dagster.core.types.decorator import dagster_type, as_dagster_type


from dagster.core.types.marshal import SerializationStrategy

from dagster.core.types.runtime import RuntimeType, define_python_dagster_type

from dagster.utils import file_relative_path
from dagster.utils.test import execute_solid, execute_solids_within_pipeline

from .version import __version__


__all__ = [
    # Definition
    'CompositeSolidDefinition',
    'DependencyDefinition',
    'EventMetadataEntry',
    'TextMetadataEntryData',
    'UrlMetadataEntryData',
    'PathMetadataEntryData',
    'JsonMetadataEntryData',
    'ExpectationResult',
    'Failure',
    'Field',
    'InputDefinition',
    'LoggerDefinition',
    'Materialization',
    'ModeDefinition',
    'OutputDefinition',
    'PipelineDefinition',
    'RepositoryDefinition',
    'ResourceDefinition',
    'Output',
    'SolidDefinition',
    'SolidInvocation',
    'SystemStorageDefinition',
    # Decorators
    'composite_solid',
    'lambda_solid',
    'logger',
    'pipeline',
    'resource',
    'solid',
    'system_storage',
    # Execution
    'execute_pipeline_iterator',
    'execute_pipeline',
    'DagsterEventType',
    'InitLoggerContext',
    'InitResourceContext',
    'InitSystemStorageContext',
    'InProcessExecutorConfig',
    'MultiprocessExecutorConfig',
    'PipelineExecutionResult',
    'RunConfig',
    'SolidExecutionResult',
    'SystemStorageData',
    # Errors
    'DagsterExecutionStepExecutionError',
    'DagsterExecutionStepNotFoundError',
    'DagsterInvalidConfigError',
    'DagsterInvalidDefinitionError',
    'DagsterInvariantViolationError',
    'DagsterResourceFunctionError',
    'DagsterTypeCheckError',
    'DagsterUserCodeExecutionError',
    'DagsterStepOutputNotFoundError',
    # Utilities
    'execute_solid',
    'execute_solids_within_pipeline',
    'file_relative_path',
    # types
    'Any',
    'Bool',
    'input_hydration_config',
    'Dict',
    'Enum',
    'EnumValue',
    'Float',
    'Int',
    'List',
    'NamedDict',
    'Optional',
    'output_materialization_config',
    'Path',
    'PermissiveDict',
    'String',
    'SerializationStrategy',
    'Nothing',
    # type creation
    'as_dagster_type',
    'dagster_type',
    'define_python_dagster_type',
    'RuntimeType',
    'ConfigType',
    'ConfigScalar',
    # file things
    'FileHandle',
    'LocalFileHandle',
]



from abc import ABCMeta, abstractmethod
from collections import namedtuple
import multiprocessing
import os
import sys

import six

from dagster import check
from dagster.utils.error import serializable_error_info_from_exc_info, SerializableErrorInfo
from dagster.utils import get_multiprocessing_context


ChildProcessStartEvent = namedtuple('ChildProcessStartEvent', 'pid')
ChildProcessDoneEvent = namedtuple('ChildProcessDoneEvent', 'pid')
ChildProcessSystemErrorEvent = namedtuple('ChildProcessSystemErrorEvent', 'pid error_info')

ChildProcessEvents = (ChildProcessStartEvent, ChildProcessDoneEvent, ChildProcessSystemErrorEvent)


# Inherit from this class in order to use this library. The object must be pickable
# and you instantiate it and pass it to execute_command_in_child_process. execute()
# is invoked in a child process
class ChildProcessCommand(six.with_metaclass(ABCMeta)):  # pylint: disable=no-init
    @abstractmethod
    def execute(self):
        pass


class ChildProcessException(Exception):
    def __init__(self, *args, **kwargs):
        super(ChildProcessException, self).__init__(*args)
        self.error_info = check.inst_param(
            kwargs.pop('error_info'), 'error_info', SerializableErrorInfo
        )


class ChildProcessCrashException(Exception):
    pass


def execute_command_in_child_process(queue, command):
    check.inst_param(command, 'command', ChildProcessCommand)

    pid = os.getpid()
    queue.put(ChildProcessStartEvent(pid=pid))
    try:
        for step_event in command.execute():
            queue.put(step_event)
        queue.put(ChildProcessDoneEvent(pid=pid))
    except Exception:  # pylint: disable=broad-except
        queue.put(
            ChildProcessSystemErrorEvent(
                pid=pid, error_info=serializable_error_info_from_exc_info(sys.exc_info())
            )
        )
    finally:
        queue.close()


TICK = 20.0 * 1.0 / 1000.0  # 20 MS

PROCESS_DEAD_AND_QUEUE_EMPTY = 'PROCESS_DEAD_AND_QUEUE_EMPTY'


def get_next_event(process, queue):
    try:
        return queue.get(block=True, timeout=TICK)
    except multiprocessing.queues.Empty:
        if not process.is_alive():
            # There is a possibility that after the last queue.get the
            # process created another event and then died. In that case
            # we want to continue draining the queue.
            try:
                return queue.get(block=False)
            except multiprocessing.queues.Empty:
                # If the queue empty we know that there are no more events
                # and that the process has died.
                return PROCESS_DEAD_AND_QUEUE_EMPTY

    return None


def execute_child_process_command(command, return_process_events=False):
    '''
    This function polls the process until it returns a valid
    item or returns PROCESS_DEAD_AND_QUEUE_EMPTY if it is in
    a state where the process has terminated and the queue is empty.

    If wait_mode is set to YIELD, it will yield None while the process is busy executing.

    Warning: if the child process is in an infinite loop. This will
    also infinitely loop.
    '''

    check.inst_param(command, 'command', ChildProcessCommand)
    check.bool_param(return_process_events, 'return_process_events')

    multiprocessing_context = get_multiprocessing_context()
    queue = multiprocessing_context.Queue()

    process = multiprocessing_context.Process(
        target=execute_command_in_child_process, args=(queue, command)
    )

    process.start()

    completed_properly = False

    while not completed_properly:
        event = get_next_event(process, queue)

        # child process is busy executing, yield so we (the parent) can continue
        # other work such as checking other child_process_commands
        if event is None:
            yield None

        if event == PROCESS_DEAD_AND_QUEUE_EMPTY:
            break

        # If we are configured to return process events by the caller,
        # yield that event to the caller
        if return_process_events and isinstance(event, ChildProcessEvents):
            yield event

        if isinstance(event, ChildProcessDoneEvent):
            completed_properly = True
        elif isinstance(event, ChildProcessSystemErrorEvent):
            raise ChildProcessException(
                'Uncaught exception in process {pid} with message "{message}" and error info {error_info}'.format(
                    pid=event.pid, message=event.error_info.message, error_info=event.error_info
                ),
                error_info=event.error_info,
            )
        elif not isinstance(event, ChildProcessEvents):
            yield event

    if not completed_properly:
        # TODO Gather up stderr and the process exit code
        raise ChildProcessCrashException()

    process.join()
